# cython: language_level=3, embedsignature=True
# distutils: language=c++
# Library imports.
import numpy as np
cimport numpy as np

# Cython imports
from cython.operator cimport dereference as deref, preincrement as inc
from cython.parallel import parallel, prange, threadid

# malloc and friends
from libc.stdlib cimport malloc, free
from libcpp.map cimport map
from libcpp.pair cimport pair
from libcpp.vector cimport vector

cdef extern from "<algorithm>" namespace "std" nogil:
    void sort[Iter, Compare](Iter first, Iter last, Compare comp)
    void sort[Iter](Iter first, Iter last)

# cpython
from cpython.dict cimport PyDict_Clear, PyDict_Contains, PyDict_GetItem
from cpython.list cimport PyList_GetItem, PyList_SetItem, PyList_GET_ITEM

# Cython for compiler directives
cimport cython

from compyle.config import get_config
from compyle.array import get_backend, Array
from compyle.parallel import Elementwise
from compyle.types import annotate


IF OPENMP:
    cimport openmp
    cpdef int get_number_of_threads():
        cdef int i, n
        with nogil, parallel():
            for i in prange(1):
                n = openmp.omp_get_num_threads()
        return n
    cpdef set_number_of_threads(int n):
        openmp.omp_set_num_threads(n)
ELSE:
    cpdef int get_number_of_threads():
        return 1
    cpdef set_number_of_threads(int n):
        print("OpenMP not available, cannot set number of threads.")


IF UNAME_SYSNAME == "Windows":
    cdef inline double fmin(double x, double y) nogil:
        return x if x < y else y
    cdef inline double fmax(double x, double y) nogil:
        return x if x > y else y

# Particle Tag information
from cyarray.carray cimport BaseArray, aligned_malloc, aligned_free
from .utils import ParticleTAGS

cdef int Local = ParticleTAGS.Local
cdef int Remote = ParticleTAGS.Remote
cdef int Ghost = ParticleTAGS.Ghost

ctypedef pair[unsigned int, unsigned int] id_gid_pair_t


cdef inline bint _compare_gids(id_gid_pair_t x, id_gid_pair_t y) nogil:
    return y.second > x.second

def py_flatten(IntPoint cid, IntArray ncells_per_dim, int dim):
    """Python wrapper"""
    cdef cIntPoint _cid = cid.data
    cdef int flattened_index = flatten( _cid, ncells_per_dim, dim )
    return flattened_index

def py_get_valid_cell_index(IntPoint cid, IntArray ncells_per_dim, int dim,
                            int n_cells):
    """Return the flattened cell index for a valid cell"""
    return get_valid_cell_index(cid.data.x, cid.data.y, cid.data.z,
            ncells_per_dim.data, dim, n_cells)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline cIntPoint unflatten(long cell_index, IntArray ncells_per_dim,
                                int dim):
    """Un-flatten a linear cell index"""
    cdef int ncx = ncells_per_dim.data[0]
    cdef int ncy = ncells_per_dim.data[1]

    cdef cIntPoint cid
    cdef int ix=0, iy=0, iz=0, tmp=0

    if dim > 1:
        if dim > 2:
            tmp = ncx * ncy

            iz = cell_index/tmp
            cell_index = cell_index - iz * tmp


        iy = cell_index/ncx
        ix = cell_index - (iy * ncx)

    else:
        ix = cell_index

    # return the tuple cell index
    cid = cIntPoint_new( ix, iy, iz )
    return cid

def py_unflatten(long cell_index, IntArray ncells_per_dim, int dim):
    """Python wrapper"""
    cdef cIntPoint _cid = unflatten( cell_index, ncells_per_dim, dim )
    cdef IntPoint cid = IntPoint_from_cIntPoint(_cid)
    return cid

cdef cIntPoint find_cell_id(cPoint pnt, double cell_size):
    """ Find the cell index for the corresponding point

    Parameters
    ----------
    pnt -- the point for which the index is sought
    cell_size -- the cell size to use
    id -- output parameter holding the cell index


    Notes
    -----
    Performs a box sort based on the point and cell size

    Uses the function  `real_to_int`

    """
    cdef cIntPoint p = cIntPoint(real_to_int(pnt.x, cell_size),
                                 real_to_int(pnt.y, cell_size),
                                 real_to_int(pnt.z, cell_size))
    return p

cdef inline cPoint _get_centroid(double cell_size, cIntPoint cid):
    """ Get the centroid of the cell.

    Parameters
    ----------

    cell_size : double (input)
    Cell size used for binning

    cid : cPoint (input)
    Spatial index for a cell

    Returns
    -------

    centroid : cPoint

    Notes
    -----
    The centroid in any coordinate direction is defined to be the
    origin plus half the cell size in that direction

    """
    centroid = cPoint_new(0.0, 0.0, 0.0)
    centroid.x = (<double>cid.x + 0.5)*cell_size
    centroid.y = (<double>cid.y + 0.5)*cell_size
    centroid.z = (<double>cid.z + 0.5)*cell_size

    return centroid

def get_centroid(double cell_size, IntPoint cid):
    """ Get the centroid of the cell.

    Parameters
    ----------

    cell_size : double (input)
        Cell size used for binning

    cid : IntPoint (input)
        Spatial index for a cell

    Returns
    -------

    centroid : Point

    Notes
    -----
    The centroid in any coordinate direction is defined to be the
    origin plus half the cell size in that direction

    """
    cdef cPoint _centroid = _get_centroid(cell_size, cid.data)
    centroid = Point_new(0.0, 0.0, 0.0)

    centroid.data = _centroid
    return centroid

cpdef UIntArray arange_uint(int start, int stop=-1):
    """Utility function to return a numpy.arange for a UIntArray"""
    cdef int size
    cdef UIntArray arange
    cdef int i = 0

    if stop == -1:
        arange = UIntArray(start)
        for i in range(start):
            arange.data[i] = <unsigned int>i
    else:
        size = stop-start
        arange = UIntArray(size)
        for i in range(size):
            arange.data[i] = <unsigned int>(start + i)

    return arange

##############################################################################
cdef class NNPSParticleArrayWrapper:
    def __init__(self, ParticleArray pa):
        self.pa = pa
        self.name = pa.name

        self.x = pa.get_carray('x')
        self.y = pa.get_carray('y')
        self.z = pa.get_carray('z')
        self.h = pa.get_carray('h')

        self.gid = pa.get_carray('gid')
        self.tag = pa.get_carray('tag')

        self.np = pa.get_number_of_particles()

    cdef int get_number_of_particles(self):
        cdef ParticleArray pa = self.pa
        return pa.get_number_of_particles()

    def remove_tagged_particles(self, int tag):
        cdef ParticleArray pa = self.pa
        pa.remove_tagged_particles(tag)

##############################################################################
cdef class DomainManager:
    def __init__(self, double xmin=-1000, double xmax=1000, double ymin=0,
                 double ymax=0, double zmin=0, double zmax=0,
                 periodic_in_x=False, periodic_in_y=False, periodic_in_z=False,
                 double n_layers=2.0, backend=None, props=None,
                 mirror_in_x=False, mirror_in_y=False, mirror_in_z=False):

        """Constructor

        Parameters
        ----------

        xmin, xmax, ymin, ymax, zmin, zmax: double: extents of the domain.
        periodic_in_x, periodic_in_y, periodic_in_z: bool: axis periodicity.
        mirror_in_x, mirror_in_y, mirror_in_z: bool: axis mirror.

        n_layers: double: number of ghost layers as a multiple of
            h_max*radius_scale

        props: list/dict: properties to copy.
            Provide a list or dict with the keys as particle array names.
            Only the specified properties are copied.  If not specified,
            all props are copied.
        """
        self.backend = get_backend(backend)
        is_periodic = periodic_in_x or periodic_in_y or periodic_in_z
        is_mirror = mirror_in_x or mirror_in_y or mirror_in_z
        if (self.backend is 'opencl' or self.backend is 'cuda'):
            if not is_mirror:
                from pysph.base.gpu_domain_manager import GPUDomainManager
                domain_manager = GPUDomainManager
            else:
                print("warning: mirrored boundary conditions not "
                "supported with GPU backend, using CPUDomainManager")
                domain_manager = CPUDomainManager
        else:
            domain_manager = CPUDomainManager
        self.manager = domain_manager(
            xmin=xmin, xmax=xmax, ymin=ymin,
            ymax=ymax, zmin=zmin, zmax=zmax, periodic_in_x=periodic_in_x,
            periodic_in_y=periodic_in_y, periodic_in_z=periodic_in_z,
            n_layers=n_layers, backend=self.backend, props=props,
            mirror_in_x=mirror_in_x, mirror_in_y=mirror_in_y,
            mirror_in_z=mirror_in_z
        )

    def set_pa_wrappers(self, wrappers):
        self.manager.set_pa_wrappers(wrappers)

    def set_cell_size(self, cell_size):
        self.manager.set_cell_size(cell_size)

    def set_in_parallel(self, in_parallel):
        self.manager.set_in_parallel(in_parallel)

    def set_radius_scale(self, radius_scale):
        self.manager.set_radius_scale(radius_scale)

    def compute_cell_size_for_binning(self):
        """Compute the cell size for the binning.

        The cell size is chosen as the kernel radius scale times the
        maximum smoothing length in the local processor. For parallel
        runs, we would need to communicate the maximum 'h' on all
        processors to decide on the appropriate binning size.

        """
        self.manager.compute_cell_size_for_binning()

    def update(self):
        self.manager.update()


##############################################################################
cdef class DomainManagerBase(object):
    def __init__(self, double xmin=-1000, double xmax=1000, double ymin=0,
                 double ymax=0, double zmin=0, double zmax=0,
                 periodic_in_x=False, periodic_in_y=False, periodic_in_z=False,
                 double n_layers=2.0, props=None, mirror_in_x=False,
                 mirror_in_y=False, mirror_in_z=False):
        """Constructor

        The n_layers argument specifies the number of ghost layers as multiples
        of hmax*radius_scale.

        props: list/dict: properties to copy.
            Provide a list or dict with the keys as particle array names.
            Only the specified properties are copied.  If not specified,
            all props are copied.
        """
        self._check_limits(xmin, xmax, ymin, ymax, zmin, zmax)

        self.xmin = xmin; self.xmax = xmax
        self.ymin = ymin; self.ymax = ymax
        self.zmin = zmin; self.zmax = zmax

        self.props = props
        # Indicates if the domain is periodic
        self.periodic_in_x = periodic_in_x
        self.periodic_in_y = periodic_in_y
        self.periodic_in_z = periodic_in_z

        # Indicates if the domain is mirrored
        self.mirror_in_x = mirror_in_x
        self.mirror_in_y = mirror_in_y
        self.mirror_in_z = mirror_in_z

        self.is_periodic = periodic_in_x or periodic_in_y or periodic_in_z
        self.is_mirror = mirror_in_x or mirror_in_y or mirror_in_z
        self.n_layers = n_layers

        # get the translates in each coordinate direction
        self.xtranslate = xmax - xmin
        self.ytranslate = ymax - ymin
        self.ztranslate = zmax - zmin

        # empty list of particle array wrappers for now
        self.pa_wrappers = []
        self.narrays = 0

        # default value for the cell size
        self.cell_size = 1.0
        self.hmin = 1.0

        # default DomainManager in_parallel is set to False
        self.in_parallel = False

    def _check_limits(self, xmin, xmax, ymin, ymax, zmin, zmax):
        """Sanity check on the limits"""
        if ( (xmax < xmin) or (ymax < ymin) or (zmax < zmin) ):
            raise ValueError("Invalid domain limits!")

    def set_pa_wrappers(self, wrappers):
        self.pa_wrappers = wrappers
        self.narrays = len(wrappers)
        copy_props = []
        props = self.props
        for i in range(self.narrays):
            if props is None:
                copy_props.append(None)
            elif isinstance(props, dict):
                name = wrappers[i].pa.name
                copy_props.append(props[name])
            else:
                copy_props.append(props)
        self.copy_props = copy_props

    def set_cell_size(self, cell_size):
        self.cell_size = cell_size

    def set_in_parallel(self, bint in_parallel):
        self.in_parallel = in_parallel

    def set_radius_scale(self, double radius_scale):
        self.radius_scale = radius_scale

    def compute_cell_size_for_binning(self):
        self._compute_cell_size_for_binning()

    cpdef _remove_ghosts(self):
        """Remove all ghost particles from a previous step

        While creating periodic neighbors, we create new particles and
        give them the tag utils.ParticleTAGS.Ghost. Before repeating
        this step in the next iteration, all current particles with
        this tag are removed.

        """
        cdef list pa_wrappers = self.pa_wrappers
        cdef int narrays = self.narrays

        cdef int array_index
        cdef NNPSParticleArrayWrapper pa_wrapper

        for array_index in range(narrays):
            pa_wrapper = <NNPSParticleArrayWrapper> pa_wrappers[array_index]
            pa_wrapper.remove_tagged_particles(Ghost)


##############################################################################
cdef class CPUDomainManager(DomainManagerBase):
    """This class determines the limits of the solution domain.

    We expect all simulations to have well defined domain limits
    beyond which we are either not interested or the solution is
    invalid to begin with. Thus, if a particle leaves the domain,
    the solution should be considered invalid (at least locally).

    The initial domain limits could be given explicitly or asked to be
    computed from the particle arrays. The domain could be periodic.

    """
    def __init__(self, double xmin=-1000, double xmax=1000, double ymin=0,
                 double ymax=0, double zmin=0, double zmax=0,
                 periodic_in_x=False, periodic_in_y=False, periodic_in_z=False,
                 double n_layers=2.0, backend=None, props=None,
                 mirror_in_x=False, mirror_in_y=False, mirror_in_z=False):

        """Constructor

        The n_layers argument specifies the number of ghost layers as multiples
        of hmax*radius_scale.

        props: list/dict: properties to copy.
            Provide a list or dict with the keys as particle array names.
            Only the specified properties are copied.  If not specified,
            all props are copied.
        """
        DomainManagerBase.__init__(
            self, xmin=xmin, xmax=xmax,
            ymin=ymin, ymax=ymax, zmin=zmin, zmax=zmax,
            periodic_in_x=periodic_in_x, periodic_in_y=periodic_in_y,
            periodic_in_z=periodic_in_z, n_layers=n_layers, props=props,
            mirror_in_x=mirror_in_x, mirror_in_y=mirror_in_y,
            mirror_in_z=mirror_in_z
        )

        self.use_double = True
        self.dtype = float
        self.dtype_max = np.finfo(self.dtype).max
        self.ghosts = None

    #### Public protocol ################################################
    def update(self):
        """General method that is called before NNPS can bin particles.

        This method is responsible for the computation of cell sizes
        and creation of any ghost particles for periodic or wall
        boundary conditions.

        """
        # compute the cell sizes
        self._compute_cell_size_for_binning()

        # Periodicity is handled by adjusting particles according to a
        # given cubic domain box. In parallel, it is expected that the
        # appropriate parallel NNPS is responsible for the creation of
        # ghost particles.
        if (self.is_periodic or self.is_mirror) and not self.in_parallel:
            self._update_from_gpu()

            # remove periodic/mirror ghost particles from a previous step
            self._remove_ghosts()

            if self.is_periodic:
                # box-wrap current particles for periodicity
                self._box_wrap_periodic()

                # create new periodic ghosts
                self._create_ghosts_periodic()

            if self.is_mirror:
                # create new mirrored ghosts
                self._create_ghosts_mirror()

            # Update GPU.
            self._update_gpu()

    #### Private protocol ###############################################
    cdef _add_to_array(self, DoubleArray arr, double disp, int start=0):
        cdef int i
        for i in range(arr.length - start):
            arr.data[start + i] += disp

    cdef _change_velocity(self, DoubleArray arr, double disp):
        cdef int i
        for i in range(arr.length):
            arr.data[i] += -1.0 * disp

    cdef _add_array_to_array(self, DoubleArray arr, DoubleArray translate):
        cdef int i
        for i in range(arr.length):
            arr.data[i] += translate.data[i]

    cdef _mul_to_array(self, DoubleArray arr, double val):
        cdef int i
        for i in range(arr.length):
            arr.data[i] *= val

    cdef _create_ghosts_mirror(self):
        """Identify boundary particles and create images.

        We need to find all particles that are within a specified
        distance from the boundaries and place image copies on the
        other side of the boundary. Corner reflections need to be
        accounted for when using domains with multiple periodicity.

        The periodic domain is specified using the DomainManager object

        """
        cdef list pa_wrappers = self.pa_wrappers
        cdef int narrays = self.narrays

        # cell size used to check for periodic ghosts. For summation density
        # like operations, we need to create two layers of ghost images, this
        # is configurable via the n_layers argument to the constructor.
        cdef double cell_size = self.n_layers * self.cell_size

        # mirror domain values
        cdef double xmin = self.xmin, xmax = self.xmax
        cdef double ymin = self.ymin, ymax = self.ymax,
        cdef double zmin = self.zmin, zmax = self.zmax

        # mirror boundary condition flags
        cdef bint mirror_in_x = self.mirror_in_x
        cdef bint mirror_in_y = self.mirror_in_y
        cdef bint mirror_in_z = self.mirror_in_z

        # locals
        cdef NNPSParticleArrayWrapper pa_wrapper
        cdef ParticleArray pa, added
        cdef DoubleArray x, y, z
        cdef double xi, yi, zi
        cdef int array_index, i, np

        # temporary indices for particles to be replicated
        cdef LongArray x_low, x_high, y_high, y_low, z_high, z_low, low, high

        x_low = LongArray(); x_high = LongArray()
        y_high = LongArray(); y_low = LongArray()
        z_high = LongArray(); z_low = LongArray()

        xt_low = DoubleArray(); xt_high = DoubleArray()
        yt_high = DoubleArray(); yt_low = DoubleArray()
        zt_high = DoubleArray(); zt_low = DoubleArray()
        low = LongArray(); high = LongArray()
        low_translate = DoubleArray(); high_translate = DoubleArray()

        for array_index in range(narrays):
            pa_wrapper = pa_wrappers[ array_index ]
            pa = pa_wrapper.pa
            x = pa_wrapper.x; y = pa_wrapper.y; z = pa_wrapper.z

            # reset the length of the arrays
            x_low.reset(); x_high.reset(); y_high.reset(); y_low.reset()
            z_low.reset(); z_high.reset()

            np = x.length
            for i in range(np):
                xi = x.data[i]; yi = y.data[i]; zi = z.data[i]

                if mirror_in_x:
                    if ((xi - xmin) <= cell_size):
                        x_low.append(i)
                        xt_low.append(-2*(xi - xmin))
                    if ( (xmax - xi) <= cell_size):
                        x_high.append(i)
                        xt_high.append(2*(xmax - xi))

                if mirror_in_y:
                    if ((yi - ymin) <= cell_size):
                        y_low.append(i)
                        yt_low.append(-2*(yi - ymin))
                    if ( (ymax - yi) <= cell_size ):
                        y_high.append(i)
                        yt_high.append(2*(ymax - yi))

                if mirror_in_z:
                    if ((zi - zmin) <= cell_size):
                        z_low.append(i)
                        zt_low.append(-2*(zi - zmin))
                    if ((zmax - zi) <= cell_size ):
                        z_high.append(i)
                        zt_high.append(2*(zmax - zi))


            # now treat each case separately and append to the main array
            added = ParticleArray(x=None, y=None, z=None)
            x = added.get_carray('x')
            y = added.get_carray('y')
            z = added.get_carray('z')
            if mirror_in_x:
                # x_low
                copy = pa.extract_particles( x_low )
                if copy.get_number_of_particles() > 0:
                    self._add_array_to_array(copy.get_carray('x'), xt_low)
                    self._mul_to_array(copy.get_carray('u'), -1)
                    added.append_parray(copy)

                # x_high
                copy = pa.extract_particles( x_high )
                if copy.get_number_of_particles() > 0:
                    self._add_array_to_array(copy.get_carray('x'), xt_high)
                    self._mul_to_array(copy.get_carray('u'), -1)
                    added.append_parray(copy)

            if mirror_in_y:
                # Now do the corners from the previous.
                low.reset(); high.reset()
                low_translate.reset(); high_translate.reset()
                np = x.length
                for i in range(np):
                    yi = y.data[i]
                    if ( (yi - ymin) <= cell_size ):
                        low.append(i)
                        low_translate.append(-2*(yi - ymin))
                    if ( (ymax - yi) <= cell_size ):
                        high.append(i)
                        high_translate.append(2*(ymax - yi))

                copy = added.extract_particles(low)
                if copy.get_number_of_particles() > 0:
                    self._add_array_to_array(copy.get_carray('y'), low_translate)
                    self._mul_to_array(copy.get_carray('v'), -1)
                    added.append_parray(copy)

                copy = added.extract_particles(high)
                if copy.get_number_of_particles() > 0:
                    self._add_array_to_array(copy.get_carray('y'), high_translate)
                    self._mul_to_array(copy.get_carray('v'), -1)
                    added.append_parray(copy)

                # Add the actual y_high and y_low now.
                # y_high
                copy = pa.extract_particles( y_high )
                if copy.get_number_of_particles() > 0:
                    self._add_array_to_array(copy.get_carray('y'), yt_high)
                    self._mul_to_array(copy.get_carray('v'), -1)
                    added.append_parray(copy)

                # y_low
                copy = pa.extract_particles( y_low )
                if copy.get_number_of_particles() > 0:
                    self._add_array_to_array(copy.get_carray('y'), yt_low)
                    self._mul_to_array(copy.get_carray('v'), -1)
                    added.append_parray(copy)

            if mirror_in_z:
                # Now do the corners from the previous.
                low.reset(); high.reset()
                low_translate.reset(); high_translate.reset()
                np = x.length
                for i in range(np):
                    zi = z.data[i]
                    if ((zi - zmin) <= cell_size):
                        low.append(i)
                        low_translate.append(-2*(zi - zmin))
                    if ((zmax - zi) <= cell_size):
                        high.append(i)
                        high_translate.append(2*(zmax - zi))

                copy = added.extract_particles(low)
                if copy.get_number_of_particles() > 0:
                    self._add_array_to_array(copy.get_carray('z'), low_translate)
                    self._mul_to_array(copy.get_carray('w'), -1)
                    added.append_parray(copy)

                copy = added.extract_particles(high)
                if copy.get_number_of_particles() > 0:
                    self._add_array_to_array(copy.get_carray('z'), high_translate)
                    self._mul_to_array(copy.get_carray('w'), -1)
                    added.append_parray(copy)

                # Add the actual z_high and z_low now.
                # z_high
                copy = pa.extract_particles( z_high )
                if copy.get_number_of_particles() > 0:
                    self._add_array_to_array(copy.get_carray('z'), zt_high)
                    self._mul_to_array(copy.get_carray('w'), -1)
                    added.append_parray(copy)

                # z_low
                copy = pa.extract_particles( z_low )
                if copy.get_number_of_particles() > 0:
                    self._add_array_to_array(copy.get_carray('z'), zt_low)
                    self._mul_to_array(copy.get_carray('w'), -1)
                    added.append_parray(copy)


            added.tag[:] = Ghost
            pa.append_parray(added)

    cdef _box_wrap_periodic(self):
        """Box-wrap particles for periodicity

        The periodic domain is a rectangular box defined by minimum
        and maximum values in each coordinate direction. These values
        are used in turn to define translation values used to box-wrap
        particles that cross a periodic boundary.

        The periodic domain is specified using the DomainManager object

        """
        # minimum and maximum values of the domain
        cdef double xmin = self.xmin, xmax = self.xmax
        cdef double ymin = self.ymin, ymax = self.ymax
        cdef double zmin = self.zmin, zmax = self.zmax

        # translations along each coordinate direction
        cdef double xtranslate = self.xtranslate
        cdef double ytranslate = self.ytranslate
        cdef double ztranslate = self.ztranslate

        # periodicity flags for NNPS
        cdef bint periodic_in_x = self.periodic_in_x
        cdef bint periodic_in_y = self.periodic_in_y
        cdef bint periodic_in_z = self.periodic_in_z

        # locals
        cdef NNPSParticleArrayWrapper pa_wrapper
        cdef DoubleArray x, y, z
        cdef double xi, yi, zi
        cdef int i, np

        # iterate over each array and mark for translation
        for pa_wrapper in self.pa_wrappers:
            x = pa_wrapper.x; y = pa_wrapper.y; z = pa_wrapper.z
            np = x.length

            # iterate over particles and box-wrap
            for i in range(np):

                if periodic_in_x:
                    if x.data[i] < xmin : x.data[i] = x.data[i] + xtranslate
                    if x.data[i] > xmax : x.data[i] = x.data[i] - xtranslate

                if periodic_in_y:
                    if y.data[i] < ymin : y.data[i] = y.data[i] + ytranslate
                    if y.data[i] > ymax : y.data[i] = y.data[i] - ytranslate

                if periodic_in_z:
                    if z.data[i] < zmin : z.data[i] = z.data[i] + ztranslate
                    if z.data[i] > zmax : z.data[i] = z.data[i] - ztranslate

    cdef _create_ghosts_periodic(self):
        """Identify boundary particles and create images.

        We need to find all particles that are within a specified
        distance from the boundaries and place image copies on the
        other side of the boundary. Corner reflections need to be
        accounted for when using domains with multiple periodicity.

        The periodic domain is specified using the DomainManager object

        """
        cdef list pa_wrappers = self.pa_wrappers
        cdef int narrays = self.narrays
        cdef list copy_props = self.copy_props

        # cell size used to check for periodic ghosts. For summation density
        # like operations, we need to create two layers of ghost images, this
        # is configurable via the n_layers argument to the constructor.
        cdef double cell_size = self.n_layers * self.cell_size

        # periodic domain values
        cdef double xmin = self.xmin, xmax = self.xmax
        cdef double ymin = self.ymin, ymax = self.ymax,
        cdef double zmin = self.zmin, zmax = self.zmax

        cdef double xtranslate = self.xtranslate
        cdef double ytranslate = self.ytranslate
        cdef double ztranslate = self.ztranslate

        # periodicity flags
        cdef bint periodic_in_x = self.periodic_in_x
        cdef bint periodic_in_y = self.periodic_in_y
        cdef bint periodic_in_z = self.periodic_in_z

        # locals
        cdef NNPSParticleArrayWrapper pa_wrapper
        cdef ParticleArray pa, ghost_pa
        cdef DoubleArray x, y, z, xt, yt, zt
        cdef double xi, yi, zi
        cdef int array_index, i, np, start

        # temporary indices for particles to be replicated
        cdef LongArray x_low, x_high, y_high, y_low, z_high, z_low, low, high

        x_low = LongArray(); x_high = LongArray()
        y_high = LongArray(); y_low = LongArray()
        z_high = LongArray(); z_low = LongArray()
        low = LongArray(); high = LongArray()

        if not self.ghosts:
            self.ghosts = [paw.pa.empty_clone(props=copy_props[i])
                           for i, paw in enumerate(pa_wrappers)]
        else:
            for ghost_pa in self.ghosts:
                ghost_pa.resize(0)
            for i in range(narrays):
                self.ghosts[i].ensure_properties(
                    pa_wrappers[i].pa, props=copy_props[i]
                )

        for array_index in range(narrays):
            ghost_pa = self.ghosts[array_index]
            pa_wrapper = pa_wrappers[array_index]
            pa = pa_wrapper.pa
            x = pa_wrapper.x; y = pa_wrapper.y; z = pa_wrapper.z

            # reset the length of the arrays
            x_low.reset(); x_high.reset(); y_high.reset(); y_low.reset()
            z_low.reset(); z_high.reset()

            np = x.length
            for i in range(np):
                xi = x.data[i]; yi = y.data[i]; zi = z.data[i]

                if periodic_in_x:
                    if ( (xi - xmin) <= cell_size ): x_low.append(i)
                    if ( (xmax - xi) <= cell_size ): x_high.append(i)

                if periodic_in_y:
                    if ( (yi - ymin) <= cell_size ): y_low.append(i)
                    if ( (ymax - yi) <= cell_size ): y_high.append(i)

                if periodic_in_z:
                    if ( (zi - zmin) <= cell_size ): z_low.append(i)
                    if ( (zmax - zi) <= cell_size ): z_high.append(i)

            # now treat each case separately and append to the main array
            x = ghost_pa.get_carray('x')
            y = ghost_pa.get_carray('y')
            z = ghost_pa.get_carray('z')
            if periodic_in_x:
                # x_low
                start = ghost_pa.get_number_of_particles()
                pa.extract_particles(
                    x_low, ghost_pa, align=False, props=copy_props[array_index]
                )
                self._add_to_array(ghost_pa.get_carray('x'), xtranslate,
                                   start=start)

                # x_high
                start = ghost_pa.get_number_of_particles()
                pa.extract_particles(
                    x_high, ghost_pa, align=False, props=copy_props[array_index]
                )
                self._add_to_array(ghost_pa.get_carray('x'), -xtranslate,
                                   start=start)

            if periodic_in_y:
                # Now do the corners from the previous.
                low.reset(); high.reset()
                np = x.length
                for i in range(np):
                    yi = y.data[i]
                    if ( (yi - ymin) <= cell_size ): low.append(i)
                    if ( (ymax - yi) <= cell_size ): high.append(i)

                start = ghost_pa.get_number_of_particles()
                ghost_pa.extract_particles(
                    low, ghost_pa, align=False, props=copy_props[array_index]
                )
                self._add_to_array(ghost_pa.get_carray('y'), ytranslate,
                                   start=start)

                start = ghost_pa.get_number_of_particles()
                ghost_pa.extract_particles(
                    high, ghost_pa, align=False, props=copy_props[array_index]
                )
                self._add_to_array(ghost_pa.get_carray('y'), -ytranslate,
                                   start=start)

                # Add the actual y_high and y_low now.
                # y_high
                start = ghost_pa.get_number_of_particles()
                pa.extract_particles(
                    y_high, ghost_pa, align=False, props=copy_props[array_index]
                )
                self._add_to_array(ghost_pa.get_carray('y'), -ytranslate,
                                   start=start)

                # y_low
                start = ghost_pa.get_number_of_particles()
                pa.extract_particles(
                    y_low, ghost_pa, align=False, props=copy_props[array_index]
                )
                self._add_to_array(ghost_pa.get_carray('y'), ytranslate,
                                   start=start)

            if periodic_in_z:
                # Now do the corners from the previous.
                low.reset(); high.reset()
                np = x.length
                for i in range(np):
                    zi = z.data[i]
                    if ( (zi - zmin) <= cell_size ): low.append(i)
                    if ( (zmax - zi) <= cell_size ): high.append(i)

                start = ghost_pa.get_number_of_particles()
                ghost_pa.extract_particles(
                    low, ghost_pa, align=False, props=copy_props[array_index]
                )
                self._add_to_array(ghost_pa.get_carray('z'), ztranslate,
                                   start=start)

                start = ghost_pa.get_number_of_particles()
                ghost_pa.extract_particles(
                    high, ghost_pa, align=False, props=copy_props[array_index]
                )
                self._add_to_array(ghost_pa.get_carray('z'), -ztranslate,
                                   start=start)

                # Add the actual z_high and z_low now.
                # z_high
                start = ghost_pa.get_number_of_particles()
                pa.extract_particles(
                    z_high, ghost_pa, align=False, props=copy_props[array_index]
                )
                self._add_to_array(ghost_pa.get_carray('z'), -ztranslate,
                                   start=start)

                # z_low
                start = ghost_pa.get_number_of_particles()
                pa.extract_particles(
                    z_low, ghost_pa, align=False, props=copy_props[array_index]
                )
                self._add_to_array(ghost_pa.get_carray('z'), ztranslate,
                                   start=start)

            ghost_pa.set_num_real_particles(ghost_pa.get_number_of_particles())
            ghost_pa.tag[:] = Ghost
            pa.append_parray(ghost_pa, align=False)

    cdef _compute_cell_size_for_binning(self):
        """Compute the cell size for the binning.

        The cell size is chosen as the kernel radius scale times the
        maximum smoothing length in the local processor. For parallel
        runs, we would need to communicate the maximum 'h' on all
        processors to decide on the appropriate binning size.

        """
        cdef list pa_wrappers = self.pa_wrappers
        cdef NNPSParticleArrayWrapper pa_wrapper
        cdef DoubleArray h
        cdef double cell_size
        cdef double _hmax, hmax = -1.0
        cdef double _hmin, hmin = self.dtype_max

        for pa_wrapper in pa_wrappers:
            h = pa_wrapper.h
            h.update_min_max()

            _hmax = h.maximum
            _hmin = h.minimum
            if _hmax > hmax:
                hmax = _hmax
            if _hmin < hmin:
                hmin = _hmin

        cell_size = self.radius_scale * hmax
        self.hmin = self.radius_scale * hmin

        if cell_size < 1e-6:
            cell_size = 1.0

        self.cell_size = cell_size

        # set the cell size for the DomainManager
        self.set_cell_size(cell_size)

    def _update_gpu(self):
        # FIXME: this is just done for correctness.  We should really
        # implement a GPU Domain Manager and use that instead to do all
        # this directly on the GPU.
        cdef list pa_wrappers = self.pa_wrappers
        cdef NNPSParticleArrayWrapper pa_wrapper
        cdef ParticleArray pa
        cdef list props

        for pa_wrapper in pa_wrappers:
            pa = pa_wrapper.pa
            if pa.gpu is not None:
                props = pa.output_property_arrays
                if len(props) == 0:
                    props = list(pa.properties.keys())
                pa.gpu.resize(pa.get_number_of_particles(real=False))
                pa.gpu.push(*props)

    def _update_from_gpu(self):
        # FIXME: this is just done for correctness.  We should really
        # implement a GPU Domain Manager and use that instead to do all
        # this directly on the GPU.
        cdef list pa_wrappers = self.pa_wrappers
        cdef NNPSParticleArrayWrapper pa_wrapper
        cdef ParticleArray pa
        cdef list props

        for pa_wrapper in pa_wrappers:
            pa = pa_wrapper.pa
            if pa.gpu is not None:
                props = pa.output_property_arrays
                if len(props) == 0:
                    props = list(pa.properties.keys())
                pa.gpu.pull(*props)


##############################################################################
cdef class Cell:
    """Basic indexing structure for the box-sort NNPS.

    For a spatial indexing based on the box-sort algorithm, this class
    defines the spatial data structure used to hold particle indices
    (local and global) that are within this cell.

    """
    def __init__(self, IntPoint cid, double cell_size, int narrays,
                 int layers=2):
        """Constructor

        Parameters
        ----------

        cid : IntPoint
            Spatial index (unflattened) for the cell

        cell_size : double
            Spatial extent of the cell in each dimension

        narrays : int
            Number of arrays being binned

        layers : int
            Factor to compute the bounding box

        """
        self._cid = cIntPoint_new(cid.x, cid.y, cid.z)
        self.cell_size = cell_size

        self.narrays = narrays

        self.lindices = [UIntArray() for i in range(narrays)]
        self.gindices = [UIntArray() for i in range(narrays)]

        self.nparticles = [lindices.length for lindices in self.lindices]
        self.is_boundary = False

        # compute the centroid for the cell
        self.centroid = _get_centroid(cell_size, cid.data)

        # cell bounding box
        self.layers = layers
        self._compute_bounding_box(cell_size, layers)

        # list of neighboring processors
        self.nbrprocs = IntArray(0)

        # current size of the cell
        self.size = 0.0

    #### Public protocol ################################################
    def get_centroid(self, Point pnt):
        """Utility function to get the centroid of the cell.

        Parameters
        ----------

        pnt : Point (input/output)
            The centroid is cmoputed and stored in this object.

        The centroid is defined as the origin plus half the cell size
        in each dimension.

        """
        cdef cPoint centroid = self.centroid
        pnt.data = centroid

    def get_bounding_box(self, Point boxmin, Point boxmax, int layers = 1,
                         cell_size=None):
        """Compute the bounding box for the cell.

        Parameters
        ----------

        boxmin : Point (output)
            The bounding box min coordinates are stored here

        boxmax : Point (output)
            The bounding box max coordinates are stored here

        layers : int (input) default (1)
            Number of offset layers to define the bounding box

        cell_size : double (input) default (None)
            Optional cell size to use to compute the bounding box.
            If not provided, the cell's size will be used.

        """
        if cell_size is None:
            cell_size = self.cell_size

        self._compute_bounding_box(cell_size, layers)
        boxmin.data = self.boxmin
        boxmax.data = self.boxmax

    cpdef set_indices(self, int index, UIntArray lindices, UIntArray gindices):
        """Set the global and local indices for the cell"""
        self.lindices[index] = lindices
        self.gindices[index] = gindices
        self.nparticles[index] = lindices.length

    #### Private protocol ###############################################

    cdef _compute_bounding_box(self, double cell_size,
                               int layers):
        self.layers = layers
        cdef cPoint centroid = self.centroid
        cdef cPoint boxmin = cPoint_new(0., 0., 0.)
        cdef cPoint boxmax = cPoint_new(0., 0., 0.)

        boxmin.x = centroid.x - (layers+0.5) * cell_size
        boxmax.x = centroid.x + (layers+0.5) * cell_size

        boxmin.y = centroid.y - (layers+0.5) * cell_size
        boxmax.y = centroid.y + (layers+0.5) * cell_size

        boxmin.z = centroid.z - (layers + 0.5) * cell_size
        boxmax.z = centroid.z + (layers + 0.5) * cell_size

        self.boxmin = boxmin
        self.boxmax = boxmax


###############################################################################

cdef class NeighborCache:
    def __init__(self, NNPS nnps, int dst_index, int src_index):
        self._dst_index = dst_index
        self._src_index = src_index
        self._nnps = nnps
        self._particles = nnps.particles
        self._narrays = nnps.narrays

        cdef long n_p = self._particles[dst_index].get_number_of_particles()
        cdef int nnbr = 10
        cdef size_t i
        if self._nnps.dim == 1:
            nnbr = 10
        elif self._nnps.dim == 2:
            nnbr = 60
        elif self._nnps.dim == 3:
            nnbr = 120

        self._n_threads = get_number_of_threads()

        self._cached = IntArray(n_p)
        for i in range(n_p):
            self._cached.data[i] = 0

        self._last_avg_nbr_size = nnbr
        self._start_stop = UIntArray()
        self._pid_to_tid = UIntArray()
        self._neighbor_arrays = []
        self._neighbors = <void**>aligned_malloc(
            sizeof(void*)*self._n_threads
        )

        cdef UIntArray _arr
        for i in range(self._n_threads):
            _arr = UIntArray()
            self._neighbor_arrays.append(_arr)
            self._neighbors[i] = <void*>_arr

    def __dealloc__(self):
        aligned_free(self._neighbors)

    #### Public protocol ################################################

    cdef void get_neighbors_raw(self, size_t d_idx, UIntArray nbrs) nogil:
        if self._cached.data[d_idx] == 0:
            self._find_neighbors(d_idx)
        cdef size_t start, end, tid
        start = self._start_stop.data[2*d_idx]
        end = self._start_stop.data[2*d_idx + 1]
        tid = self._pid_to_tid.data[d_idx]
        nbrs.c_set_view(
            &(<UIntArray>self._neighbors[tid]).data[start], end - start
        )

    cpdef get_neighbors(self, int src_index, size_t d_idx, UIntArray nbrs):
        self.get_neighbors_raw(d_idx, nbrs)

    cpdef find_all_neighbors(self):
        cdef long d_idx
        cdef long np = \
                self._particles[self._dst_index].get_number_of_particles()

        with nogil, parallel():
            for d_idx in prange(np):
                if self._cached.data[d_idx] == 0:
                    self._find_neighbors(d_idx)

    cpdef update(self):
        self._update_last_avg_nbr_size()
        cdef int n_threads = self._n_threads
        cdef int dst_index = self._dst_index
        cdef size_t i
        cdef long np = self._particles[dst_index].get_number_of_particles()
        self._start_stop.resize(np*2)
        self._pid_to_tid.resize(np)
        self._cached.resize(np)
        for i in range(np):
            self._cached.data[i] = 0
            self._start_stop.data[2*i] = 0
            self._start_stop.data[2*i+1] = 0
        # This is an upper limit for the number of neighbors in a worst
        # case scenario.
        cdef size_t safety = 1024
        cdef UIntArray arr
        for i in range(n_threads):
            arr = <UIntArray>self._neighbors[i]
            arr.c_reset()
            arr.c_reserve(
                <size_t>(self._last_avg_nbr_size*np/n_threads) + safety
            )

    #### Private protocol ################################################

    cdef void _update_last_avg_nbr_size(self):
        cdef size_t i
        cdef size_t np = self._pid_to_tid.length
        cdef UIntArray start_stop = self._start_stop
        cdef long total = 0
        for i in range(np):
            total += start_stop.data[2*i + 1] - start_stop.data[2*i]
        if total > 0 and np > 0:
            self._last_avg_nbr_size = int(total/np) + 1

    cdef void _find_neighbors(self, long d_idx) nogil:
        cdef int thread_id = threadid()
        self._pid_to_tid.data[d_idx] = thread_id
        self._start_stop.data[d_idx*2] = \
            (<UIntArray>self._neighbors[thread_id]).length
        self._nnps.find_nearest_neighbors(
            d_idx, <UIntArray>self._neighbors[thread_id]
        )
        self._start_stop.data[d_idx*2+1] = \
            (<UIntArray>self._neighbors[thread_id]).length
        self._cached.data[d_idx] = 1


##############################################################################
cdef class NNPSBase:
    def __init__(self, int dim, list particles, double radius_scale=2.0,
                 int ghost_layers=1, domain=None, bint cache=False,
                 bint sort_gids=False):
        """Constructor for NNPS

        Parameters
        ----------

        dim : int
            Dimension (fixme: Not sure if this is really needed)

        particles : list
            The list of particle arrays we are working on.

        radius_scale : double, default (2)
            Optional kernel radius scale. Defaults to 2

        domain : DomainManager, default (None)
            Optional limits for the domain

        cache : bint
            Flag to set if we want to cache neighbor calls. This costs
            storage but speeds up neighbor calculations.

        sort_gids : bint, default (False)
            Flag to sort neighbors using gids (if they are available).
            This is useful when comparing parallel results with those
            from a serial run.
        """
        # store the list of particles and number of arrays
        self.particles = particles
        self.narrays = len( particles )

        # create the particle array wrappers
        self.pa_wrappers = [NNPSParticleArrayWrapper(pa) for pa in particles]

        # radius scale and problem dimensionality.
        self.radius_scale = radius_scale
        self.dim = dim

        self.domain = domain
        if domain is None:
            self.domain = DomainManager()

        # set the particle array wrappers for the domain manager
        self.domain.set_pa_wrappers(self.pa_wrappers)

        # set the radius scale to determine the cell size
        self.domain.set_radius_scale(self.radius_scale)

        # periodicity
        self.is_periodic = self.domain.manager.is_periodic

        # The total number of cells.
        self.n_cells = 0

        # cell shifts. Indicates the shifting value for cell indices
        # in each co-ordinate direction.
        self.cell_shifts = IntArray(3)
        self.cell_shifts.data[0] = -1
        self.cell_shifts.data[1] = 0
        self.cell_shifts.data[2] = 1

    cpdef brute_force_neighbors(self, int src_index, int dst_index,
                                size_t d_idx, UIntArray nbrs):
        cdef NNPSParticleArrayWrapper src = self.pa_wrappers[src_index]
        cdef NNPSParticleArrayWrapper dst = self.pa_wrappers[dst_index]

        cdef DoubleArray s_x = src.x
        cdef DoubleArray s_y = src.y
        cdef DoubleArray s_z = src.z
        cdef DoubleArray s_h = src.h

        cdef DoubleArray d_x = dst.x
        cdef DoubleArray d_y = dst.y
        cdef DoubleArray d_z = dst.z
        cdef DoubleArray d_h = dst.h

        cdef double cell_size = self.cell_size
        cdef double radius_scale = self.radius_scale

        cdef size_t num_particles, j

        num_particles = s_x.length
        cdef double xi = d_x.data[d_idx]
        cdef double yi = d_y.data[d_idx]
        cdef double zi = d_z.data[d_idx]

        cdef double hi = d_h.data[d_idx] * radius_scale # gather radius
        cdef double xj, yj, hj, xij2, xij

        # reset the neighbors
        nbrs.reset()

        for j in range(num_particles):
            xj = s_x.data[j]; yj = s_y.data[j]; zj = s_z.data[j];
            hj = radius_scale * s_h.data[j] # scatter radius

            xij2 = (xi - xj)*(xi - xj) + \
                   (yi - yj)*(yi - yj) + \
                   (zi - zj)*(zi - zj)
            xij = sqrt(xij2)

            if ( (xij < hi) or (xij < hj) ):
                nbrs.append( <ZOLTAN_ID_TYPE> j )

    cpdef get_nearest_particles_no_cache(self, int src_index, int dst_index,
            size_t d_idx, UIntArray nbrs, bint prealloc):
        """Find nearest neighbors for particle id 'd_idx' without cache

        Parameters
        ----------
        src_index: int
            Index in the list of particle arrays to which the neighbors belong

        dst_index: int
            Index in the list of particle arrays to which the query point belongs

        d_idx: size_t
            Index of the query point in the destination particle array

        nbrs: UIntArray
            Array to be populated by nearest neighbors of 'd_idx'

        """
        self.set_context(src_index, dst_index)

        if prealloc:
            nbrs.length = 0
        else:
            nbrs.c_reset()

        self.find_nearest_neighbors(d_idx, nbrs)

    cdef void find_nearest_neighbors(self, size_t d_idx, UIntArray nbrs) nogil:
        # Implement this in the subclass to actually do something useful.
        pass


    cpdef get_nearest_particles(self, int src_index, int dst_index,
                                size_t d_idx, UIntArray nbrs):
        cdef int idx = dst_index*self.narrays + src_index
        if self.use_cache:
            if self.src_index != src_index \
                or self.dst_index != dst_index:
                self.set_context(src_index, dst_index)
            return self.cache[idx].get_neighbors(src_index, d_idx, nbrs)
        else:
            return self.get_nearest_particles_no_cache(
                src_index, dst_index, d_idx, nbrs, False
            )

    cpdef set_context(self, int src_index, int dst_index):
        """Setup the context before asking for neighbors.  The `dst_index`
        represents the particles for whom the neighbors are to be determined
        from the particle array with index `src_index`.

        Parameters
        ----------

         src_index: int: the source index of the particle array.
         dst_index: int: the destination index of the particle array.
        """
        cdef int idx = dst_index*self.narrays + src_index
        self.src_index = src_index
        self.dst_index = dst_index
        self.current_cache = self.cache[idx]

cdef class NNPS(NNPSBase):
    """Nearest neighbor query class using the box-sort algorithm.

    NNPS bins all local particles using the box sort algorithm in
    Cells. The cells are stored in a dictionary 'cells' which is keyed
    on the spatial index (IntPoint) of the cell.

    """
    def __init__(self, int dim, list particles, double radius_scale=2.0,
                 int ghost_layers=1, domain=None, bint cache=False,
                 bint sort_gids=False):
        NNPSBase.__init__(self, dim, particles, radius_scale, ghost_layers,
                domain, cache, sort_gids)

        # min and max coordinate values
        self.xmin = DoubleArray(3)
        self.xmax = DoubleArray(3)
        self._last_domain_size = 0.0

        # The cache.
        self.use_cache = cache
        _cache = []
        for d_idx in range(len(particles)):
            for s_idx in range(len(particles)):
                _cache.append(NeighborCache(self, d_idx, s_idx))
        self.cache = _cache

    #### Public protocol #################################################

    def set_in_parallel(self, bint in_parallel):
        self.domain.manager.in_parallel = in_parallel

    def set_use_cache(self, bint use_cache):
        self.use_cache = use_cache
        if use_cache:
            for cache in self.cache:
                cache.update()

    def update_domain(self):
        self.domain.update()

    cpdef update(self):
        """Update the local data after particles have moved.

        For parallel runs, we want the NNPS to be independent of the
        ParallelManager which is solely responsible for distributing
        particles across available processors. We assume therefore
        that after a parallel update, each processor has all the local
        particle information it needs and this operation is carried
        out locally.

        For serial runs, this method should be called when the
        particles have moved.

        """
        cdef int i, num_particles
        cdef ParticleArray pa
        cdef UIntArray indices

        cdef DomainManager domain = self.domain

        # use cell sizes computed by the domain.
        self.cell_size = domain.manager.cell_size
        self.hmin = domain.manager.hmin

        # compute bounds and refresh the data structure
        self._compute_bounds()
        self._refresh()

        # indices on which to bin. We bin all local particles
        for i in range(self.narrays):
            pa = self.particles[i]
            num_particles = pa.get_number_of_particles()
            indices = arange_uint(num_particles)

            # bin the particles
            self._bin( pa_index=i, indices=indices )

        if self.use_cache:
            for cache in self.cache:
                cache.update()

    cdef void get_nearest_neighbors(self, size_t d_idx, UIntArray nbrs) nogil:
        if self.use_cache:
            self.current_cache.get_neighbors_raw(d_idx, nbrs)
        else:
            nbrs.c_reset()
            self.find_nearest_neighbors(d_idx, nbrs)

    #### Private protocol ################################################
    cdef _compute_bounds(self):
        """Compute coordinate bounds for the particles"""
        cdef list pa_wrappers = self.pa_wrappers
        cdef NNPSParticleArrayWrapper pa_wrapper
        cdef DoubleArray x, y, z
        cdef double xmax = -1e100, ymax = -1e100, zmax = -1e100
        cdef double xmin = 1e100, ymin = 1e100, zmin = 1e100
        cdef double lx, ly, lz, domain_size

        for pa_wrapper in pa_wrappers:
            x = pa_wrapper.x
            y = pa_wrapper.y
            z = pa_wrapper.z

            # find min and max of variables
            x.update_min_max()
            y.update_min_max()
            z.update_min_max()

            xmax = fmax(x.maximum, xmax)
            ymax = fmax(y.maximum, ymax)
            zmax = fmax(z.maximum, zmax)

            xmin = fmin(x.minimum, xmin)
            ymin = fmin(y.minimum, ymin)
            zmin = fmin(z.minimum, zmin)

        # Add a small offset to the limits.
        lx, ly, lz = xmax - xmin, ymax - ymin, zmax - zmin
        xmin -= lx*0.01; ymin -= ly*0.01; zmin -= lz*0.01
        xmax += lx*0.01; ymax += ly*0.01; zmax += lz*0.01

        domain_size = fmax(lx, ly)
        domain_size = fmax(domain_size, lz)
        if self._last_domain_size > 1e-16 and \
           domain_size > 2.0*self._last_domain_size:
            msg = (
                '*'*70 +
                '\nWARNING: Domain size has increased by a large amount.\n' +
                'Particles are probably diverging, please check your code!\n' +
                '*'*70
            )
            print(msg)
        self._last_domain_size = domain_size

        # If all of the dimensions have very small extent give it a unit size.
        cdef double _eps = 1e-12
        if (fabs(xmax - xmin) < _eps) and (fabs(ymax - ymin) < _eps) \
            and (fabs(zmax - zmin) < _eps):
            xmin -= 0.5; xmax += 0.5
            ymin -= 0.5; ymax += 0.5
            zmin -= 0.5; zmax += 0.5

        # store the minimum and maximum of physical coordinates
        self.xmin.set_data(np.asarray([xmin, ymin, zmin]))
        self.xmax.set_data(np.asarray([xmax, ymax, zmax]))

    cdef void _sort_neighbors(self, unsigned int* nbrs, size_t length,
                              unsigned int *gids) nogil:
        if length == 0:
            return
        cdef id_gid_pair_t _entry
        cdef vector[id_gid_pair_t] _data
        cdef vector[unsigned int] _ids
        cdef int i
        cdef unsigned int _id

        if gids[0] == UINT_MAX:
            # Serial runs will have invalid gids so just compare the ids.
            _ids.resize(length)
            for i in range(length):
                _ids[i] = nbrs[i]
            sort(_ids.begin(), _ids.end())
            for i in range(length):
                nbrs[i] = _ids[i]
        else:
            # Copy the neighbor id and gid data.
            _data.resize(length)
            for i in range(length):
                _id = nbrs[i]
                _entry.first = _id
                _entry.second = gids[_id]
                _data[i] = _entry
            # Sort it.
            sort(_data.begin(), _data.end(), &_compare_gids)
            # Set the sorted neighbors.
            for i in range(length):
                nbrs[i] = _data[i].first

    cpdef _bin(self, int pa_index, UIntArray indices):
        raise NotImplementedError("NNPS :: _bin called")

    cpdef _refresh(self):
        raise NotImplementedError("NNPS :: _refresh called")

    cpdef get_spatially_ordered_indices(self, int pa_index, LongArray indices):
        raise NotImplementedError("NNPSBase :: get_spatially_ordered_indices called")

    cpdef spatially_order_particles(self, int pa_index):
        """Spatially order particles such that nearby particles have indices
        nearer each other.  This may improve pre-fetching on the CPU.
        """
        cdef LongArray indices = LongArray()
        cdef ParticleArray pa = self.pa_wrappers[pa_index].pa
        self.get_spatially_ordered_indices(pa_index, indices)
        cdef BaseArray arr

        for name, arr in pa.properties.items():
            stride = pa.stride.get(name, 1)
            arr.c_align_array(indices, stride)



from argparse import ArgumentParser

from pysph.sph.scheme import SchemeChooser, WCSPHScheme
from pysph.sph.wc.edac import EDACScheme


def test_scheme_chooser_does_not_clobber_default():

    # When
    wcsph = WCSPHScheme(
        ['f'], ['b'], dim=2, rho0=1.0, c0=10.0,
        h0=0.1, hdx=1.3, alpha=0.2, beta=0.1,
    )
    edac = EDACScheme(
        fluids=['f'], solids=['b'], dim=2, c0=10.0, nu=0.001,
        rho0=1.0, h=0.1, alpha=0.0, pb=0.0
    )
    s = SchemeChooser(default='wcsph', wcsph=wcsph, edac=edac)
    p = ArgumentParser(conflict_handler="resolve")
    s.add_user_options(p)
    opts = p.parse_args([])

    # When
    s.consume_user_options(opts)

    # Then
    assert s.scheme.alpha == 0.2
    assert s.scheme.beta == 0.1

    # When
    opts = p.parse_args(['--alpha', '0.3', '--beta', '0.4'])
    s.consume_user_options(opts)

    # Then
    assert s.scheme.alpha == 0.3
    assert s.scheme.beta == 0.4



from pysph.base.gpu_nnps_base import GPUNeighborCache, GPUNNPS, BruteForceNNPS
from pysph.base.z_order_gpu_nnps import ZOrderGPUNNPS
from pysph.base.stratified_sfc_gpu_nnps import StratifiedSFCGPUNNPS
from pysph.base.gpu_domain_manager import GPUDomainManager
from pysph.base.octree_gpu_nnps import OctreeGPUNNPS



'''Setup script for PySPH.

You can use some environment variables to control the build. Setting CC/CXX
will let you choose a custom compiler (these can also be set in the config file
discussed below). You can also export ZOLTAN or ZOLTAN_INCLUDE/ZOLTAN_LIBRARY.

These are more configuration file options that trump everything. The file is in
~/.compyle/config.py.  The options are:

OMP_CFLAGS = ['-fopenmp']
OMP_LINK = ['-fopenmp']

# MPI options: handy on clusters like a Cray.
MPI_CFLAGS = ['...']  # must be a list.
MPI_LINK = ['...']

# Zoltan options
USE_TRILINOS = 1  # When set to anything, use "-ltrilinos_zoltan".
ZOLTAN = '/path/to_zoltan'  # looks inside this for $ZOLTAN/include/, lib/

# Not needed if using ZOLTAN
ZOLTAN_INCLUDE = 'path/include'  # path to zoltan.h
ZOLTAN_LIBRARY = 'path/lib'  # path to libzoltan.a

'''

import os
from subprocess import check_output
import sys


# This is taken from compyle.ext_module.
def get_config_file_opts():
    '''A global configuration file is used to configure build options
    for compyle and other packages.  This is located in:

    ~/.compyle/config.py

    The file can contain arbitrary Python that is exec'd. The variables defined
    here specify the compile and link args. For example, one may set:

    OMP_CFLAGS = ['-fopenmp']
    OMP_LINK = ['-fopenmp']

    Will use these instead of the defaults that are automatically determined.
    These must be lists.

    '''
    fname = os.path.expanduser(os.path.join('~', '.compyle', 'config.py'))
    opts = {}
    if os.path.exists(fname):
        print('Reading configuration options from %s.' % fname)
        with open(fname) as fp:
            exec(compile(fp.read(), fname, 'exec'), opts)
        opts.pop('__builtins__', None)
    return opts


# NOTE: the configuration options in the file trump everything else!
# These are options from the .compyle/config.py
CONFIG_OPTS = get_config_file_opts()


if len(os.environ.get('COVERAGE', '')) > 0:
    MACROS = [("CYTHON_TRACE", "1"), ("CYTHON_TRACE_NOGIL", "1")]
    COMPILER_DIRECTIVES = {"linetrace": True}
    print("-" * 80)
    print("Enabling linetracing for cython and setting CYTHON_TRACE = 1")
    print("-" * 80)
else:
    MACROS = []
    COMPILER_DIRECTIVES = {}

MODE = 'normal'
if len(sys.argv) >= 2 and \
   ('--help' in sys.argv[1:] or
    sys.argv[1] in ('--help-commands', 'egg_info', '--version',
                    'clean', 'sdist')):
    MODE = 'info'

HAVE_MPI = True
try:
    import mpi4py
except ImportError:
    HAVE_MPI = False

USE_ZOLTAN = True
HAVE_ZOLTAN = True
try:
    import pyzoltan  # noqa: F401
except ImportError:
    HAVE_ZOLTAN = False

base_includes = [] if sys.platform == 'win32' else ['/usr/local/include/']

compiler = 'gcc'
# compiler = 'intel'
if compiler == 'intel':
    extra_compile_args = ['-O3']
else:
    extra_compile_args = []


def get_deps(*args):
    """Given a list of basenames, this checks if a .pyx or .pxd exists
    and returns the list.
    """
    result = []
    for basename in args:
        for ext in ('.pyx', '.pxd'):
            f = basename + ext
            if os.path.exists(f):
                result.append(f)
    return result


def _get_openmp_flags():
    """Return the OpenMP flags for the platform.

    This returns two lists, [extra_compile_args], [extra_link_args]
    """
    # Copied from compyle.ext_module
    if 'OMP_CFLAGS' in CONFIG_OPTS or 'OMP_LINK' in CONFIG_OPTS:
        return CONFIG_OPTS['OMP_CFLAGS'], CONFIG_OPTS['OMP_LINK']

    if sys.platform == 'win32':
        return ['/openmp'], []
    elif sys.platform == 'darwin':
        if (os.environ.get('CC') is not None and
           os.environ.get('CXX') is not None):
            return ['-fopenmp'], ['-fopenmp']
        else:
            return ['-Xpreprocessor', '-fopenmp'], ['-lomp']
    else:
        return ['-fopenmp'], ['-fopenmp']


def get_openmp_flags():
    """Returns any OpenMP related flags if OpenMP is avaiable on the system.
    """
    omp_compile_flags, omp_link_flags = _get_openmp_flags()

    from textwrap import dedent
    try:
        from Cython.Distutils import Extension
        from pyximport import pyxbuild
    except ImportError:
        print("Unable to import Cython, disabling OpenMP for now.")
        return [], [], False

    from distutils.errors import CompileError, LinkError
    import shutil
    import tempfile
    test_code = dedent("""
    from cython.parallel import parallel, prange, threadid
    cimport openmp
    def n_threads():
        with nogil, parallel():
            openmp.omp_get_num_threads()
    """)
    tmp_dir = tempfile.mkdtemp()
    fname = os.path.join(tmp_dir, 'check_omp.pyx')
    with open(fname, 'w') as fp:
        fp.write(test_code)
    extension = Extension(
        name='check_omp', sources=[fname],
        include_dirs=base_includes,
        extra_compile_args=omp_compile_flags,
        extra_link_args=omp_link_flags,
    )
    has_omp = True
    try:
        pyxbuild.pyx_to_dll(fname, extension, pyxbuild_dir=tmp_dir)
        print("-" * 70)
        print("Using OpenMP.")
        print("-" * 70)
    except CompileError:
        print("*" * 70)
        print("Unable to compile OpenMP code. Not using OpenMP.")
        print("*" * 70)
        has_omp = False
    except LinkError:
        print("*" * 70)
        print("Unable to link OpenMP code. Not using OpenMP.")
        print("*" * 70)
        has_omp = False
    finally:
        shutil.rmtree(tmp_dir)

    if has_omp:
        return omp_compile_flags, omp_link_flags, True
    else:
        return [], [], False


def get_zoltan_directory(varname):
    global USE_ZOLTAN
    if varname in CONFIG_OPTS:
        return os.path.expanduser(CONFIG_OPTS[varname])

    d = os.environ.get(varname, '')
    if len(d) == 0:
        USE_ZOLTAN = False
        return ''
    else:
        USE_ZOLTAN = True
    if not os.path.exists(d):
        print("*" * 80)
        print("%s incorrectly set to %s, not using ZOLTAN!" % (varname, d))
        print("*" * 80)
        USE_ZOLTAN = False
        return ''
    return d


def get_mpi_flags():
    """Returns mpi_inc_dirs, mpi_compile_args, mpi_link_args.
    """
    global HAVE_MPI
    mpi_inc_dirs = []
    mpi_compile_args = []
    mpi_link_args = []
    if not HAVE_MPI:
        return mpi_inc_dirs, mpi_compile_args, mpi_link_args
    elif 'MPI_CFLAGS' in CONFIG_OPTS:
        mpi_compile_args = CONFIG_OPTS['MPI_CFLAGS']
        mpi_link_args = CONFIG_OPTS['MPI_LINK']
        mpi_inc_dirs.append(mpi4py.get_include())
    else:
        try:
            mpic = 'mpic++'
            if compiler == 'intel':
                link_args = check_output(
                    [mpic, '-cc=icc', '-link_info'],
                    universal_newlines=True
                ).strip()
                link_args = link_args[3:]
                compile_args = check_output(
                    [mpic, '-cc=icc', '-compile_info'],
                    universal_newlines=True
                ).strip()
                compile_args = compile_args[3:]
            else:
                link_args = check_output(
                    [mpic, '--showme:link'],
                    universal_newlines=True
                ).strip()
                compile_args = check_output(
                    [mpic, '--showme:compile'],
                    universal_newlines=True
                ).strip()
        except:  # noqa: E722
            print('-' * 80)
            print("Unable to run mpic++ correctly, skipping parallel build")
            print('-' * 80)
            HAVE_MPI = False
        else:
            mpi_link_args.extend(link_args.split())
            mpi_compile_args.extend(compile_args.split())
            mpi_inc_dirs.append(mpi4py.get_include())

    return mpi_inc_dirs, mpi_compile_args, mpi_link_args


def get_zoltan_args():
    """Returns zoltan_include_dirs, zoltan_library_dirs
    """
    global HAVE_MPI, USE_ZOLTAN
    zoltan_include_dirs, zoltan_library_dirs = [], []
    if not HAVE_MPI or not HAVE_ZOLTAN:
        return zoltan_include_dirs, zoltan_library_dirs
    # First try with the environment variable 'ZOLTAN'
    zoltan_base = get_zoltan_directory('ZOLTAN')
    inc = lib = ''
    if len(zoltan_base) > 0:
        inc = os.path.join(zoltan_base, 'include')
        lib = os.path.join(zoltan_base, 'lib')
        if not os.path.exists(inc) or not os.path.exists(lib):
            inc = lib = ''

    # try with the older ZOLTAN include directories
    if len(inc) == 0 or len(lib) == 0:
        inc = get_zoltan_directory('ZOLTAN_INCLUDE')
        lib = get_zoltan_directory('ZOLTAN_LIBRARY')

    if HAVE_ZOLTAN and not USE_ZOLTAN:
        # Try with default in sys.prefix/{include,lib}, this is what is done
        # by any conda installs of zoltan.
        inc = os.path.join(sys.prefix, 'include')
        lib = os.path.join(sys.prefix, 'lib')
        if os.path.exists(os.path.join(inc, 'zoltan.h')):
            USE_ZOLTAN = True

    if (not USE_ZOLTAN):
        print("*" * 80)
        print("Zoltan Environment variable not set, not using ZOLTAN!")
        print("*" * 80)
        HAVE_MPI = False
    else:
        print('-' * 70)
        print("Using Zoltan from:\n%s\n%s" % (inc, lib))
        print('-' * 70)
        zoltan_include_dirs = [inc]
        zoltan_library_dirs = [lib]

        # PyZoltan includes
        zoltan_cython_include = [
            os.path.abspath(
                os.path.join(os.path.dirname(pyzoltan.__file__), 'czoltan')
            )
        ]
        zoltan_include_dirs += zoltan_cython_include

    return zoltan_include_dirs, zoltan_library_dirs


def get_basic_extensions():
    if MODE == 'info':
        try:
            from Cython.Distutils import Extension
        except ImportError:
            from distutils.core import Extension
        try:
            import numpy
        except ImportError:
            include_dirs = []
        else:
            include_dirs = [numpy.get_include()]
    else:
        from Cython.Distutils import Extension
        import numpy
        include_dirs = [numpy.get_include()]

    include_dirs += base_includes
    openmp_compile_args, openmp_link_args, openmp_env = get_openmp_flags()

    ext_modules = [
        Extension(
            name="pysph.base.particle_array",
            sources=["pysph/base/particle_array.pyx"],
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args,
            language="c++",
            define_macros=MACROS,
        ),

        Extension(
            name="pysph.base.point",
            sources=["pysph/base/point.pyx"],
            extra_compile_args=extra_compile_args,
            include_dirs=include_dirs,
            language="c++",
            define_macros=MACROS,
        ),

        Extension(
            name="pysph.base.nnps_base",
            sources=["pysph/base/nnps_base.pyx"],
            depends=get_deps(
                "pysph/base/point",
                "pysph/base/particle_array",
            ),
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args + openmp_compile_args,
            extra_link_args=openmp_link_args,
            cython_compile_time_env={'OPENMP': openmp_env},
            language="c++",
            define_macros=MACROS,
        ),


        Extension(
            name="pysph.base.linked_list_nnps",
            sources=["pysph/base/linked_list_nnps.pyx"],
            depends=get_deps(
                "pysph/base/nnps_base"
            ),
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args + openmp_compile_args,
            extra_link_args=openmp_link_args,
            cython_compile_time_env={'OPENMP': openmp_env},
            language="c++",
            define_macros=MACROS,
        ),

        Extension(
            name="pysph.base.box_sort_nnps",
            sources=["pysph/base/box_sort_nnps.pyx"],
            depends=get_deps(
                "pysph/base/nnps_base", "pysph/base/linked_list_nnps"
            ),
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args + openmp_compile_args,
            extra_link_args=openmp_link_args,
            cython_compile_time_env={'OPENMP': openmp_env},
            language="c++",
            define_macros=MACROS,
        ),

        Extension(
            name="pysph.base.spatial_hash_nnps",
            sources=["pysph/base/spatial_hash_nnps.pyx"],
            depends=get_deps(
                "pysph/base/nnps_base"
            ),
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args + openmp_compile_args,
            extra_link_args=openmp_link_args,
            cython_compile_time_env={'OPENMP': openmp_env},
            language="c++",
            define_macros=MACROS,
        ),

        Extension(
            name="pysph.base.cell_indexing_nnps",
            sources=["pysph/base/cell_indexing_nnps.pyx"],
            depends=get_deps(
                "pysph/base/nnps_base"
            ),
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args + openmp_compile_args,
            extra_link_args=openmp_link_args,
            cython_compile_time_env={'OPENMP': openmp_env},
            language="c++",
            define_macros=MACROS,
        ),


        Extension(
            name="pysph.base.z_order_nnps",
            sources=["pysph/base/z_order_nnps.pyx"],
            depends=get_deps(
                "pysph/base/nnps_base"
            ),
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args + openmp_compile_args,
            extra_link_args=openmp_link_args,
            cython_compile_time_env={'OPENMP': openmp_env},
            language="c++",
            define_macros=MACROS,
        ),

        Extension(
            name="pysph.base.stratified_hash_nnps",
            sources=["pysph/base/stratified_hash_nnps.pyx"],
            depends=get_deps(
                "pysph/base/nnps_base"
            ),
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args + openmp_compile_args,
            extra_link_args=openmp_link_args,
            cython_compile_time_env={'OPENMP': openmp_env},
            language="c++",
            define_macros=MACROS,
        ),

        Extension(
            name="pysph.base.stratified_sfc_nnps",
            sources=["pysph/base/stratified_sfc_nnps.pyx"],
            depends=get_deps(
                "pysph/base/nnps_base", "pysph/base/z_order_nnps"
            ),
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args + openmp_compile_args,
            extra_link_args=openmp_link_args,
            cython_compile_time_env={'OPENMP': openmp_env},
            language="c++",
            define_macros=MACROS,
        ),

        Extension(
            name="pysph.base.octree",
            sources=["pysph/base/octree.pyx"],
            depends=get_deps(
                "pysph/base/nnps_base"
            ),
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args + openmp_compile_args,
            extra_link_args=openmp_link_args,
            cython_compile_time_env={'OPENMP': openmp_env},
            language="c++",
            define_macros=MACROS,
        ),

        Extension(
            name="pysph.base.octree_nnps",
            sources=["pysph/base/octree_nnps.pyx"],
            depends=get_deps(
                "pysph/base/nnps_base", "pysph/base/octree"
            ),
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args + openmp_compile_args,
            extra_link_args=openmp_link_args,
            cython_compile_time_env={'OPENMP': openmp_env},
            language="c++",
            define_macros=MACROS,
        ),

        # kernels used for tests
        Extension(
            name="pysph.base.c_kernels",
            sources=["pysph/base/c_kernels.pyx"],
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args,
            language="c++",
            define_macros=MACROS,
        ),

        # Eigen decomposition code
        Extension(
            name="pysph.base.linalg3",
            sources=["pysph/base/linalg3.pyx"],
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args,
            language="c++",
            define_macros=MACROS,
        ),

        # STL tool
        Extension(
            name="pysph.tools.mesh_tools",
            sources=["pysph/tools/mesh_tools.pyx"],
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args,
            language="c++",
            define_macros=MACROS,
        ),

    ]

    # OpenCL related modules
    ext_modules.extend((
        Extension(
            name="pysph.base.gpu_nnps_base",
            sources=["pysph/base/gpu_nnps_base.pyx"],
            depends=get_deps(
                "pysph/base/point",
                "pysph/base/particle_array", "pysph/base/nnps_base"
            ),
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args + openmp_compile_args,
            extra_link_args=openmp_link_args,
            cython_compile_time_env={'OPENMP': openmp_env},
            language="c++",
            define_macros=MACROS,
        ),

        Extension(
            name="pysph.base.z_order_gpu_nnps",
            sources=["pysph/base/z_order_gpu_nnps.pyx"],
            depends=get_deps(
                "pysph/base/nnps_base"
            ),
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args + openmp_compile_args,
            extra_link_args=openmp_link_args,
            cython_compile_time_env={'OPENMP': openmp_env},
            language="c++",
            define_macros=MACROS,
        ),
        Extension(
            name="pysph.base.stratified_sfc_gpu_nnps",
            sources=["pysph/base/stratified_sfc_gpu_nnps.pyx"],
            depends=get_deps(
                "pysph/base/nnps_base"
            ),
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args + openmp_compile_args,
            extra_link_args=openmp_link_args,
            cython_compile_time_env={'OPENMP': openmp_env},
            language="c++",
            define_macros=MACROS,
        ),
        Extension(
            name="pysph.base.octree_gpu_nnps",
            sources=["pysph/base/octree_gpu_nnps.pyx"],
            depends=get_deps(
                "pysph/base/nnps_base"
            ),
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args + openmp_compile_args,
            extra_link_args=openmp_link_args,
            cython_compile_time_env={'OPENMP': openmp_env},
            language="c++",
            define_macros=MACROS,
        ))
    )

    return ext_modules


def get_parallel_extensions():
    if not HAVE_MPI:
        return []

    if MODE == 'info':
        from distutils.core import Extension
        include_dirs = []
        mpi_inc_dirs, mpi_compile_args, mpi_link_args = [], [], []
        zoltan_include_dirs, zoltan_library_dirs = [], []
    else:
        from Cython.Distutils import Extension
        import numpy
        include_dirs = [numpy.get_include()]
        mpi_inc_dirs, mpi_compile_args, mpi_link_args = get_mpi_flags()
        zoltan_include_dirs, zoltan_library_dirs = get_zoltan_args()

    # We should check again here as HAVE_MPI may be set to False when we try to
    # get the MPI flags and are not successful.
    if not HAVE_MPI:
        return []

    include_dirs += base_includes

    MPI4PY_V2 = False if mpi4py.__version__.startswith('1.') else True
    cython_compile_time_env = {'MPI4PY_V2': MPI4PY_V2}

    zoltan_lib = 'zoltan'
    if 'USE_TRILINOS' in CONFIG_OPTS:
        zoltan_lib = 'trilinos_zoltan'
    elif os.environ.get('USE_TRILINOS', None) is not None:
        zoltan_lib = 'trilinos_zoltan'

    parallel_modules = [
        Extension(
            name="pysph.parallel.parallel_manager",
            sources=["pysph/parallel/parallel_manager.pyx"],
            depends=get_deps(
                "pysph/base/point", "pysph/base/particle_array",
                "pysph/base/nnps_base"
            ),
            include_dirs=include_dirs + mpi_inc_dirs + zoltan_include_dirs,
            library_dirs=zoltan_library_dirs,
            libraries=[zoltan_lib, 'mpi'],
            extra_link_args=mpi_link_args,
            extra_compile_args=mpi_compile_args + extra_compile_args,
            cython_compile_time_env=cython_compile_time_env,
            language="c++",
            define_macros=MACROS,
        ),
    ]
    return parallel_modules


def create_sources():
    argv = sys.argv
    if 'build_ext' in argv or 'develop' in sys.argv or 'install' in argv:
        pth = os.path.join('pysph', 'base')
        cmd = [sys.executable, '-m', 'cyarray.generator', os.path.abspath(pth)]
        print(check_output(cmd).decode())


def _is_cythonize_default():
    import warnings
    result = True
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        try:
            # old_build_ext was introduced in Cython 0.25 and this is when
            # cythonize was made the default.
            from Cython.Distutils import old_build_ext  # noqa: F401
        except ImportError:
            result = False
    return result


def setup_package():
    from setuptools import find_packages, setup
    if MODE == 'info':
        cmdclass = {}
    else:
        from Cython.Distutils import build_ext
        cmdclass = {'build_ext': build_ext}

        create_sources()

    # Extract the version information from pysph/__init__.py
    info = {}
    module = os.path.join('pysph', '__init__.py')
    exec(compile(open(module).read(), module, 'exec'), info)

    # The requirements.
    install_requires = [
        'numpy', 'mako', 'cyarray', 'compyle>=0.8', 'Cython>=0.20',
        'setuptools>=42.0.0', 'pytools', 'Beaker'
    ]
    tests_require = ['pytest>=3.0', 'h5py']
    if sys.version_info[:2] == (2, 6):
        install_requires += [
            'ordereddict', 'importlib'
        ]
        tests_require += ['unittest2']
    if sys.version_info[0] < 3:
        tests_require += [
            'mock>=1.0'
        ]
    docs_require = ["sphinx"]

    extras_require = dict(
        mpi=['mpi4py>=1.2', 'pyzoltan'],
        opencl=['pyopencl'],
        ui=['mayavi>=4.0', 'pyside2', 'h5py'],
        tests=tests_require,
        docs=docs_require,
        dev=tests_require + docs_require,
    )

    everything = set()
    for dep in extras_require.values():
        everything.update(dep)
    extras_require['all'] = everything

    ext_modules = get_basic_extensions() + get_parallel_extensions()
    if MODE != 'info' and _is_cythonize_default():
        # Cython >= 0.25 uses cythonize to compile the extensions. This
        # requires the compile_time_env to be set explicitly to work.
        compile_env = {}
        include_path = set()
        for mod in ext_modules:
            compile_env.update(mod.cython_compile_time_env or {})
            include_path.update(mod.include_dirs)
        from Cython.Build import cythonize
        ext_modules = cythonize(
            ext_modules, compile_time_env=compile_env,
            include_path=list(include_path),
            language="c++",
            compiler_directives=COMPILER_DIRECTIVES,
        )

    setup(name='PySPH',
          version=info['__version__'],
          author='PySPH Developers',
          author_email='pysph-dev@googlegroups.com',
          description="A general purpose Smoothed Particle Hydrodynamics "
          "framework",
          long_description=open('README.rst').read(),
          url='http://github.com/pypr/pysph',
          license="BSD",
          keywords="SPH simulation computational fluid dynamics",
          packages=find_packages(),
          package_data={
              '': ['*.pxd', '*.mako', '*.txt.gz', '*.txt', '*.txt.bz2',
                   '*.vtk.gz', '*.gz', '*.csv',
                   '*.rst', 'ndspmhd-sedov-initial-conditions.npz']
          },
          # exclude package data in installation.
          exclude_package_data={
              '': ['Makefile', '*.bat', '*.cfg', '*.rst', '*.sh', '*.yml'],
          },
          ext_modules=ext_modules,
          include_package_data=True,
          cmdclass=cmdclass,
          install_requires=install_requires,
          extras_require=extras_require,
          zip_safe=False,
          entry_points="""
              [console_scripts]
              pysph = pysph.tools.cli:main
              """,
          platforms=['Linux', 'Mac OS-X', 'Unix', 'Windows'],
          classifiers=[c.strip() for c in """\
            Development Status :: 4 - Beta
            Environment :: Console
            Intended Audience :: Developers
            Intended Audience :: Science/Research
            License :: OSI Approved :: BSD License
            Natural Language :: English
            Operating System :: MacOS :: MacOS X
            Operating System :: Microsoft :: Windows
            Operating System :: POSIX
            Operating System :: Unix
            Programming Language :: Python
            Programming Language :: Python :: 3
            Topic :: Scientific/Engineering
            Topic :: Scientific/Engineering :: Physics
            Topic :: Software Development :: Libraries
            """.splitlines() if len(c.split()) > 0],
          )


if __name__ == '__main__':
    setup_package()



"""
Hybrid Outlet proposed by

    - Negi, Pawan, Prabhu Ramachandran, and Asmelash Haftu. "An improved
      non-reflecting outlet boundary condition for weakly-compressible SPH."
      arXiv preprint arXiv:1907.04034 (2019).

"""

from pysph.sph.equation import Equation
from pysph.sph.integrator import PECIntegrator
from compyle.api import declare
import numpy

from pysph.sph.wc.edac import EDACScheme
from pysph.sph.bc.inlet_outlet_manager import InletOutletManager


class SimpleInletOutlet(InletOutletManager):
    def add_io_properties(self, pa, scheme=None):
        N = 6
        DEFAULT_PROPS = [
            'disp', 'ioid', 'Bp', 'xn', 'yn', 'zn', 'A', 'wij',
            'po', 'uho', 'vho', 'who', 'Buh', 'Bvh', 'Bwh', 'x0', 'y0', 'z0',
            'uhat', 'vhat', 'what', 'uag', 'vag', 'pag', 'pacu', 'uacu', 'uta',
            'pta', 'Eacu', 'vo', 'uo', 'wo', 'J1', 'J2u']
        STRIDE_DATA = {
            'A': 16, 'Bu': 4, 'Bv': 4, 'Bw': 4, 'Bp': 4, 'Brho': 4, 'uo':
            4, 'vo': 4, 'wo': 4, 'po': 4, 'rhoo': 4, 'Bau': 4, 'Bav': 4,
            'Baw': 4, 'auo': 4, 'avo': 4, 'awo': 4, 'Buh': 4, 'Bvh': 4,
            'Bwh': 4, 'uho': 4, 'vho': 4, 'who': 4, 'Bauh': 4, 'Bavh': 4,
            'Bawh': 4, 'auho': 4, 'avho': 4, 'awho': 4, 'Baz': 4, 'axo': 4,
            'ayo': 4, 'Bay': 4, 'azo': 4, 'Bax': 4, 'uag': N, 'vag': N,
            'pag': N, 'uhag': N, 'vhag': N}
        for prop in DEFAULT_PROPS:
            if prop in STRIDE_DATA:
                pa.add_property(prop, stride=STRIDE_DATA[prop])
            else:
                pa.add_property(prop)

        pa.add_constant('avgj2u', 0.0)
        pa.add_constant('avgj1', 0.0)
        pa.add_constant('uref', 0.0)

    def get_stepper(self, scheme, cls, edactvf=True):
        from pysph.sph.bc.inlet_outlet_manager import InletStep, OutletStep
        steppers = {}
        if (cls == PECIntegrator):
            if isinstance(scheme, EDACScheme):
                for inlet in self.inlets:
                    steppers[inlet] = InletStep()
                for outlet in self.outlets:
                    steppers[outlet] = OutletStep()
                self.active_stages = [2]

        return steppers

    def get_equations(self, scheme=None, summation_density=False,
                      edactvf=True):
        from pysph.sph.equation import Group
        from pysph.sph.bc.interpolate import (
            UpdateMomentMatrix, EvaluateUhat, EvaluateP, ExtrapolateUhat,
            ExtrapolateP, CopyUhatFromGhost, CopyPFromGhost)
        from pysph.sph.bc.inlet_outlet_manager import (
            UpdateNormalsAndDisplacements, CopyNormalsandDistances)

        all_pairs = {**self.inlet_pairs, **self.outlet_pairs}

        umax = []
        for info in self.inletinfo:
            umax.append(info.umax)

        equations = []
        g00 = []
        i = 0
        for info in self.inletinfo:
            g00.append(UpdateNormalsAndDisplacements(
                dest=info.pa_name, sources=None, xn=info.normal[0],
                yn=info.normal[1], zn=info.normal[2], xo=info.refpoint[0],
                yo=info.refpoint[1], zo=info.refpoint[2]
                ))
            g00.append(CopyNormalsandDistances(
                dest=all_pairs[info.pa_name], sources=[info.pa_name]))
            i = i + 1

        equations.append(Group(equations=g00, real=False))

        g02 = []
        for name in self.fluids:
            g02.append(CopyTimeValues(dest=name, sources=None, rho=scheme.rho0,
                                      c0=scheme.c0, u0=min(umax)))
            g02.append((EvalauteCharacterisctics(dest=name, sources=None,
                                                 c0=scheme.c0,
                                                 rho0=scheme.rho0)))

        for name in self.ghost_inlets:
            g02.append(UpdateMomentMatrix(
                dest=name, sources=self.fluids, dim=self.dim))

        equations.append(Group(equations=g02, real=False))

        g02a = []
        for name in self.fluids:
            g02a.append(ComputeTimeAverage(dest=name, sources=None))
        for name in self.outlets:
            g02a.append(EvalauteNumberdensity(dest=name, sources=self.fluids))
            g02a.append(ShepardInterpolateCharacteristics(dest=name,
                                                          sources=self.fluids))
        equations.append(Group(equations=g02a, real=False))

        g03 = []
        for name in self.ghost_inlets:
            g03.append(EvaluateUhat(dest=name, sources=self.fluids,
                                    dim=self.dim))
            g03.append(EvaluateP(dest=name, sources=self.fluids,
                                 dim=self.dim))
        equations.append(Group(equations=g03, real=False))

        g04 = []
        for name in self.ghost_inlets:
            g04.append(ExtrapolateUhat(dest=name, sources=None))
            g04.append(ExtrapolateP(dest=name, sources=None))
        for name in self.outlets:
            g04.append(EvaluatePropertyfromCharacteristics(
                dest=name, sources=None, c0=scheme.c0, rho0=scheme.rho0))
        equations.append(Group(equations=g04, real=False))

        g05 = []
        for io in self.inlet_pairs.keys():
            g05.append(CopyUhatFromGhost(
                dest=io, sources=[all_pairs[io]]))
            g05.append(CopyPFromGhost(
                dest=io, sources=[all_pairs[io]]))

        equations.append(Group(equations=g05, real=False))

        g07 = []
        for inlet in self.inletinfo:
            for eqn in inlet.equations:
                g07.append(eqn)
        for outlet in self.outletinfo:
            for eqn in outlet.equations:
                g07.append(eqn)

        equations.append(Group(equations=g07, real=False))

        g08 = []
        for name in self.ghost_inlets:
            g08.append(MoveGhostInlet(dest=name, sources=None))

        equations.append(Group(equations=g08, real=False))

        return equations


class MoveGhostInlet(Equation):
    def loop(self, d_idx, d_u, d_x, dt):
        d_x[d_idx] += d_u[d_idx] * dt


class CopyTimeValues(Equation):
    def __init__(self, dest, sources, rho, c0, u0):
        self.rho = rho
        self.c0 = c0
        self.u0 = u0
        self.Imin = 0.5 * rho * u0**2

        super(CopyTimeValues, self).__init__(dest, sources)

    def initialize(self, d_idx, d_u, d_v, d_p, d_uag, d_pag, d_uta, d_pta,
                   d_Eacu, t, d_uref):
        i6, i, N = declare('int', 3)
        N = 6
        i6 = N * d_idx
        N -= 1

        for i in range(N):
            d_uag[i6+(N-i)] = d_uag[i6+(N-(i+1))]
            d_pag[i6+(N-i)] = d_pag[i6+(N-(i+1))]

        u0 = d_uref[0]
        fac = 1.0 / (2. * self.rho * self.c0)
        Imin = (0.5 * self.rho * u0**2)**2 * fac
        d_Eacu[d_idx] = d_p[d_idx] * d_p[d_idx] * fac

        if d_Eacu[d_idx] < Imin:
            d_uag[i6] = d_u[d_idx]
            d_pag[i6] = d_p[d_idx]


class ComputeTimeAverage(Equation):
    def initialize(self, d_idx, d_uag, d_pag, d_uta, d_pta):
        i6, i, N = declare('int', 3)
        N = 6
        i6 = N * d_idx

        d_uta[d_idx] = 0.0
        d_pta[d_idx] = 0.0

        for i in range(N):
            d_uta[d_idx] += d_uag[i6+i]
            d_pta[d_idx] += d_pag[i6+i]

        d_uta[d_idx] /= N
        d_pta[d_idx] /= N


class EvalauteCharacterisctics(Equation):
    def __init__(self, dest, sources, c0, rho0):
        self.c0 = c0
        self.rho0 = rho0

        super(EvalauteCharacterisctics, self).__init__(dest, sources)

    def initialize(self, d_idx, d_u, d_p, d_J1, d_J2u, d_uta, d_pta):
        a = self.c0
        uref = d_uta[d_idx]
        pref = d_pta[d_idx]

        d_J1[d_idx] = (d_p[d_idx] - pref)
        d_J2u[d_idx] = self.rho0 * a * (d_u[d_idx] - uref) + (d_p[d_idx] -
                                                              pref)


class EvalauteNumberdensity(Equation):
    def initialize(self, d_idx, d_wij):
        d_wij[d_idx] = 0.0

    def loop(self, d_idx, d_wij, WIJ):
        d_wij[d_idx] += WIJ


class ShepardInterpolateCharacteristics(Equation):
    def initialize(self, d_idx, d_J1, d_J2u):
        d_J1[d_idx] = 0.0
        d_J2u[d_idx] = 0.0

    def loop(self, d_idx, d_J1, d_J2u, s_J1, s_J2u, WIJ, s_idx):
        d_J1[d_idx] += s_J1[s_idx] * WIJ
        d_J2u[d_idx] += s_J2u[s_idx] * WIJ

    def post_loop(self, d_idx, d_J1, d_J2u, d_wij, d_avgj2u, d_avgj1):
        if d_wij[d_idx] > 1e-14:
            d_J1[d_idx] /= d_wij[d_idx]
            d_J2u[d_idx] /= d_wij[d_idx]
        else:
            d_J2u[d_idx] = d_avgj2u[0]
            d_J1[d_idx] = d_avgj1[0]

    def reduce(self, dst, t, dt):
        dst.avgj2u[0] = numpy.average(dst.J2u[dst.wij > 0.0001])
        dst.avgj1[0] = numpy.average(dst.J1[dst.wij > 0.0001])


class EvaluatePropertyfromCharacteristics(Equation):
    def __init__(self, dest, sources, c0, rho0):
        self.c0 = c0
        self.rho0 = rho0

        super(EvaluatePropertyfromCharacteristics, self).__init__(dest,
                                                                  sources)

    def initialize(self, d_idx, d_rho, d_J2u, d_uta, d_pta, d_u, d_p, dt, t):
        if t > 20 * dt:
            j2u = d_J2u[d_idx]

            d_u[d_idx] = d_uta[d_idx] + (j2u) / (2 * self.rho0 * self.c0)
            d_p[d_idx] = d_pta[d_idx] + 0.5 * (j2u)



"""Simple tests for the Integrator steps"""

import numpy
import unittest

from pysph.base.utils import get_particle_array as gpa
from pysph.sph.integrator_step import OneStageRigidBodyStep, TwoStageRigidBodyStep

class RigidBodyMotionTestCase(unittest.TestCase):
    """Tests for linear motion.

    A particle array is subjected to acceleration along one coordinate
    direction and tested for the final acceleration, position and
    velocity.

    """
    def setUp(self):
        # create a single particle particle array
        x = numpy.array( [0.] )
        y = numpy.array( [0.] )

        additional_props = ['ax', 'ay', 'az', 'u0', 'v0', 'w0', 
                            'x0', 'y0', 'z0']
                          
        # create the particle array
        self.pa = pa = gpa(additional_props, name='square', x=x, y=y)
        
        self._set_stepper()

    def _integrate(self, final_time, dt, epec=False):
        """Integrate"""
        pa = self.pa; num_part = pa.get_number_of_particles()
        stepper = self.stepper

        current_time = 0.0
        iteration = 0

        while( current_time < final_time ):

            # initialize
            for i in range(num_part):
                stepper.initialize(
                    i, pa.x, pa.y, pa.z, pa.x0, pa.y0, pa.z0, 
                    pa.u, pa.v, pa.w, pa.u0, pa.v0, pa.w0)
            
            # update accelerations for EPEC
            if epec:
                self._update_accelerations(current_time)

            # stage 1
            for i in range( num_part ):
                stepper.stage1(
                    i, pa.x, pa.y, pa.z, pa.x0, pa.y0, pa.z0, 
                    pa.u, pa.v, pa.w, pa.u0, pa.v0, pa.w0, 
                    pa.ax, pa.ay, pa.az, dt)

            # update time
            current_time = current_time + 0.5 * dt

            # evaluate between stages
            self._update_accelerations(current_time)
            
            # call stage 2
            for i in range( num_part ):
                stepper.stage2(
                    i, pa.x, pa.y, pa.z, pa.x0, pa.y0, pa.z0, 
                    pa.u, pa.v, pa.w, pa.u0, pa.v0, pa.w0, 
                    pa.ax, pa.ay, pa.az, dt)

            # update time and iteration
            current_time = current_time + 0.5 * dt
            iteration += 1

class ConstantAccelerationTestCase(RigidBodyMotionTestCase):
    """ Constnat linear acceleration """
    def _update_accelerations(self, time):
        " Constant accelerations "
        self.pa.ax[0] = 1.0
        self.pa.ay[0] = 1.0
        self.pa.az[0] = 1.0

    def _set_stepper(self):
        # create the integrator stepper class we want to test
        self.stepper = TwoStageRigidBodyStep()

    def _test_motion(self, final_time=1.0, dt=1e-2, epec=True):
        """ Test motion for constant acceleration """
        
        # we simulate a two-stage integration with constant
        # acceleration ax = 1. Initial velocities are zero so we can
        # compare with the elementry formulae: S = 1/2 * a * t * t etc...
        self._integrate( final_time, dt, epec )

        # get the particle arrays to test
        x, y, z, u, v, w = self.pa.get('x', 'y', 'z', 'u', 'v', 'w')

        # positions S = 0 + 1/2 * at^2 = 0.5
        self.assertAlmostEqual( x[0], 0.5, 14 )
        self.assertAlmostEqual( y[0], 0.5, 14 )
        self.assertAlmostEqual( z[0], 0.5, 14 )

        # velocities v = u0 + a*t = 1.0
        self.assertAlmostEqual( u[0], 1.0, 14 )
        self.assertAlmostEqual( v[0], 1.0, 14 )
        self.assertAlmostEqual( w[0], 1.0, 14 )

    def test_one_stage(self):
        self.stepper = OneStageRigidBodyStep()
        self._test_motion()

    def test_two_stage(self):
        self.stepper = TwoStageRigidBodyStep()
        self._test_motion()

if __name__ == '__main__':
    unittest.main()



import os
import numpy as np
from pytools import memoize

import pyopencl as cl
import pyopencl.cltypes
from pyopencl.elementwise import ElementwiseKernel
from pyopencl.scan import GenericScanKernel

from compyle.opencl import get_context, get_queue, named_profile
from compyle.array import Array
from pysph.base.tree.helpers import get_vector_dtype, get_helper

NODE_KERNEL_TEMPLATE = r"""
uint node_idx = i;
%(setup)s;
int child_offset = offsets[node_idx];
if (child_offset == -1) {
    uint2 pbound = pbounds[node_idx];
    if (pbound.s0 < pbound.s1) {
        %(leaf_operation)s;
    }
} else {
    %(node_operation)s;
}
%(output_expr)s;
"""

LEAF_DFS_TEMPLATE = r"""
    /*
     * Owner of properties -
     * dst tree: cids, unique_cids
     * src tree: offsets
     */
   int cid_dst = unique_cids[i];

    /*
     * Assuming max depth of 21
     * stack_idx is also equal to current layer of tree
     * child_idx = number of children iterated through
     * idx_stack = current node
     */
    char child_stack[%(max_depth)s];
    int cid_stack[%(max_depth)s];

    char idx = 0;
    child_stack[0] = 0;
    cid_stack[0] = 1;
    char flag;
    int cid_src;
    int child_offset;

    %(setup)s;
    while (idx >= 0) {

        // Recurse to find either leaf node or invalid node
        cid_src = cid_stack[idx];

        child_offset = offsets[cid_src];
        %(common_operation)s;

        while (child_offset != -1) {
            %(node_operation)s;

            idx++;
            cid_src = child_offset;
            cid_stack[idx] = cid_src;
            child_stack[idx] = 0;
            child_offset = offsets[cid_src];
        }

        if (child_offset == -1) {
            %(leaf_operation)s;
        }

        // Recurse back to find node with a valid neighbor
        while (child_stack[idx] >= %(k)s-1 && idx >= 0)
            idx--;

        // Iterate to next neighbor
        if (idx >= 0) {
            cid_stack[idx]++;
            child_stack[idx]++;
        }
    }

    %(output_expr)s;
"""

POINT_DFS_TEMPLATE = r"""
    /*
     * Owner of properties -
     * dst tree: cids, unique_cids
     * src tree: offsets
     */
   int cid_dst = cids[i];

    /*
     * Assuming max depth of 21
     * stack_idx is also equal to current layer of tree
     * child_idx = number of children iterated through
     * idx_stack = current node
     */
    char child_stack[%(max_depth)s];
    int cid_stack[%(max_depth)s];

    char idx = 0;
    child_stack[0] = 0;
    cid_stack[0] = 1;
    char flag;
    int cid_src;
    int child_offset;

    %(setup)s;
    while (idx >= 0) {

        // Recurse to find either leaf node or invalid node
        cid_src = cid_stack[idx];

        child_offset = offsets[cid_src];
        %(common_operation)s;

        while (child_offset != -1) {
            %(node_operation)s;

            idx++;
            cid_src = child_offset;
            cid_stack[idx] = cid_src;
            child_stack[idx] = 0;
            child_offset = offsets[cid_src];
        }

        if (child_offset == -1) {
            %(leaf_operation)s;
        }

        // Recurse back to find node with a valid neighbor
        while (child_stack[idx] >= %(k)s-1 && idx >= 0)
            idx--;

        // Iterate to next neighbor
        if (idx >= 0) {
            cid_stack[idx]++;
            child_stack[idx]++;
        }
    }

    %(output_expr)s;
"""


def get_M_array_initialization(k):
    result = "uint%(k)s constant M[%(k)s]  = {" % dict(k=k)
    for i in range(k):
        result += "(uint%(k)s)(%(init_value)s)," % dict(
            k=k, init_value=','.join((np.arange(k) >= i).astype(
                int).astype(str))
        )
    result += "};"
    return result


# This segmented scan is the core of the tree construction algorithm.
# Please look at a few segmented scan examples here if you haven't already
# https://en.wikipedia.org/wiki/Segmented_scan
#
# This scan generates an array `v` of k-dimensional vectors where v[i][j]
# corresponds to how many particles are in the same node as particle `i`,
# are ordered before it and are going to lie in childs 0..j.
#
# For example, consider the first layer. In this case v[23][4] gives the
# particles numbered from 0 to 23 which are going to lie in children 0 to 4
# of the root node. The segment flag bits just let us extend this to further
# layers.
#
# The array of vectors M are just a math trick. M[j] gives a vector with j
# zeros and k - j ones (With k = 4, M[1] = {0, 1, 1, 1}). Adding this up in
# the prefix sum directly gives us the required result.
@named_profile('particle_reordering', backend='opencl')
@memoize
def _get_particle_kernel(ctx, k, args, index_code):
    return GenericScanKernel(
        ctx, get_vector_dtype('uint', k), neutral="0",
        arguments=r"""__global char *seg_flag,
                    __global uint%(k)s *prefix_sum_vector,
                    """ % dict(k=k) + args,
        input_expr="M[%(index_code)s]" % dict(index_code=index_code),
        scan_expr="(across_seg_boundary ? b : a + b)",
        is_segment_start_expr="seg_flag[i]",
        output_statement=r"""prefix_sum_vector[i]=item;""",
        preamble=get_M_array_initialization(k)
    )


# The offset of a node's child is given by:
# offset of first child in next layer + k * (number of non-leaf nodes before
# given node).
# If the node is a leaf, we set this value to be -1.
@named_profile('set_offset', backend='opencl')
@memoize
def _get_set_offset_kernel(ctx, k, leaf_size):
    return GenericScanKernel(
        ctx, np.int32, neutral="0",
        arguments=r"""__global uint2 *pbounds, __global uint *offsets,
                      __global int *leaf_count, int csum_nodes_next""",
        input_expr="(pbounds[i].s1 - pbounds[i].s0 > %(leaf_size)s)" % {
            'leaf_size': leaf_size},
        scan_expr="a + b",
        output_statement=r"""{
            offsets[i] = ((pbounds[i].s1 - pbounds[i].s0 > %(leaf_size)s) ?
                           csum_nodes_next + (%(k)s * (item - 1)) : -1);
            if (i == N - 1) { *leaf_count = (N - item); }
        }""" % {'leaf_size': leaf_size, 'k': k}
    )


# Each particle belongs to a given leaf / last layer node and this cell is
# indexed by the cid array.
# The unique_cids algorithm
# 1) unique_cids_map: unique_cids_map[i] gives the number of unique_cids
#                     before index i.
# 2) unique_cids: List of unique cids.
# 3) unique_cids_count: Number of unique cids
#
# Note that unique_cids also gives us the list of leaves / last layer nodes
# which are not empty.
@named_profile('unique_cids', backend='opencl')
@memoize
def _get_unique_cids_kernel(ctx):
    return GenericScanKernel(
        ctx, np.int32, neutral="0",
        arguments=r"""int *cids, int *unique_cids_map,
                int *unique_cids, int *unique_cids_count""",
        input_expr="(i == 0 || cids[i] != cids[i-1])",
        scan_expr="a + b",
        output_statement=r"""
            if (item != prev_item) {
                unique_cids[item - 1] = cids[i];
            }
            unique_cids_map[i] = item - 1;
            if (i == N - 1) *unique_cids_count = item;
        """
    )


# A lot of leaves are going to be empty. Not really sure if this guy is of
# any use.
@named_profile('leaves', backend='opencl')
@memoize
def _get_leaves_kernel(ctx, leaf_size):
    return GenericScanKernel(
        ctx, np.int32, neutral="0",
        arguments="int *offsets, uint2 pbounds, int *leaf_cids, "
                  "int *num_leaves",
        input_expr="(pbounds[i].s1 - pbounds[i].s0 <= %(leaf_size)s)" % dict(
            leaf_size=leaf_size
        ),
        scan_expr="a+b",
        output_statement=r"""
            if (item != prev_item) {
                leaf_cids[item - 1] = i;
            }
            if (i == N - 1) *num_leaves = item;
        """
    )


@named_profile("group_cids", backend='opencl')
@memoize
def _get_cid_groups_kernel(ctx):
    return GenericScanKernel(
        ctx, np.uint32, neutral="0",
        arguments="""int *unique_cids, uint2 *pbounds,
            int *group_cids, int *group_count, int gmin, int gmax""",
        input_expr="pass(pbounds[unique_cids[i]], gmin, gmax)",
        scan_expr="(a + b)",
        output_statement=r"""
        if (item != prev_item) {
            group_cids[item - 1] = unique_cids[i];
        }
        if (i == N - 1) *group_count = item;
        """,
        preamble="""
        char pass(uint2 pbound, int gmin, int gmax) {
            int leaf_size = pbound.s1 - pbound.s0;
            return (leaf_size > gmin && leaf_size <= gmax);
        }
        """
    )


@memoize
def tree_bottom_up(ctx, args, setup, leaf_operation, node_operation,
                   output_expr, preamble=""):
    operation = NODE_KERNEL_TEMPLATE % dict(
        setup=setup,
        leaf_operation=leaf_operation,
        node_operation=node_operation,
        output_expr=output_expr
    )

    args = ', '.join(["int *offsets, uint2 *pbounds", args])

    kernel = ElementwiseKernel(ctx, args, operation=operation,
                               preamble=preamble)

    def callable(tree, *args):
        csum_nodes = tree.total_nodes
        out = None
        for i in range(tree.depth, -1, -1):
            csum_nodes_next = csum_nodes
            csum_nodes -= tree.num_nodes[i]
            out = kernel(tree.offsets.dev, tree.pbounds.dev,
                         *args,
                         slice=slice(csum_nodes, csum_nodes_next))
        return out

    return callable


@memoize
def leaf_tree_traverse(ctx, k, args, setup, node_operation, leaf_operation,
                       output_expr, common_operation, preamble=""):
    # FIXME: variable max_depth
    operation = LEAF_DFS_TEMPLATE % dict(
        setup=setup,
        leaf_operation=leaf_operation,
        node_operation=node_operation,
        common_operation=common_operation,
        output_expr=output_expr,
        max_depth=21,
        k=k
    )

    args = ', '.join(["int *unique_cids, int *cids, int *offsets", args])

    kernel = ElementwiseKernel(
        ctx, args, operation=operation, preamble=preamble)

    def callable(tree_src, tree_dst, *args):
        return kernel(
            tree_dst.unique_cids.dev[:tree_dst.unique_cid_count],
            tree_dst.cids.dev, tree_src.offsets.dev, *args
        )

    return callable


@memoize
def point_tree_traverse(ctx, k, args, setup, node_operation, leaf_operation,
                        output_expr, common_operation, preamble=""):
    # FIXME: variable max_depth
    operation = POINT_DFS_TEMPLATE % dict(
        setup=setup,
        leaf_operation=leaf_operation,
        node_operation=node_operation,
        common_operation=common_operation,
        output_expr=output_expr,
        max_depth=21,
        k=k
    )

    args = ', '.join(["int *cids, int *offsets", args])

    kernel = ElementwiseKernel(
        ctx, args, operation=operation, preamble=preamble)

    def callable(tree_src, tree_dst, *args):
        return kernel(
            tree_dst.cids.dev, tree_src.offsets.dev, *args
        )

    return callable


class Tree(object):
    """k-ary Tree
    """

    def __init__(self, n, k=8, leaf_size=32):
        self.ctx = get_context()
        self.queue = get_queue()
        self.sorted = False
        self.main_helper = get_helper(os.path.join('tree', 'tree.mako'))

        self.initialized = False
        self.preamble = ""
        self.leaf_size = leaf_size
        self.k = k
        self.n = n
        self.sorted = False
        self.depth = 0

        self.index_function_args = []
        self.index_function_arg_ctypes = []
        self.index_function_arg_dtypes = []
        self.index_function_consts = []
        self.index_function_const_ctypes = []
        self.index_code = ""

        self.set_index_function_info()

    def set_index_function_info(self):
        raise NotImplementedError

    def get_data_args(self):
        return [getattr(self, v) for v in self.index_function_args]

    def get_index_constants(self, depth):
        raise NotImplementedError

    def _initialize_data(self):
        self.sorted = False
        num_particles = self.n
        self.pids = Array(np.uint32, n=num_particles, backend='opencl')
        self.cids = Array(np.uint32, n=num_particles, backend='opencl')
        self.cids.fill(0)

        for var, dtype in zip(self.index_function_args,
                              self.index_function_arg_dtypes):
            setattr(self, var, Array(dtype, n=num_particles, backend='opencl'))

        # Filled after tree built
        self.pbounds = None
        self.offsets = None
        self.initialized = True

    def _reinitialize_data(self):
        self.sorted = False
        num_particles = self.n
        self.pids.resize(num_particles)
        self.cids.resize(num_particles)
        self.cids.fill(0)

        for var in self.index_function_args:
            getattr(self, var).resize(num_particles)

        # Filled after tree built
        self.pbounds = None
        self.offsets = None

    def _setup_build(self):
        if not self.initialized:
            self._initialize_data()
        else:
            self._reinitialize_data()

    def _build(self, fixed_depth=None):
        self._build_tree(fixed_depth)

    ###########################################################################
    # Core construction algorithm and helper functions
    ###########################################################################

    # A little bit of manual book-keeping for temporary variables.
    # More specifically, these temporary variables would otherwise be thrown
    # away after building each layer of the tree.
    # We could instead just allocate new arrays after building each layer and
    # and let the GC take care of stuff but I'm guessing this is a
    # a better approach to save on memory
    def _create_temp_vars(self, temp_vars):
        n = self.n
        temp_vars['pids'] = Array(np.uint32, n=n, backend='opencl')
        for var, dtype in zip(self.index_function_args,
                              self.index_function_arg_dtypes):
            temp_vars[var] = Array(dtype, n=n, backend='opencl')
        temp_vars['cids'] = Array(np.uint32, n=n, backend='opencl')

    def _exchange_temp_vars(self, temp_vars):
        for k in temp_vars.keys():
            t = temp_vars[k]
            temp_vars[k] = getattr(self, k)
            setattr(self, k, t)

    def _clean_temp_vars(self, temp_vars):
        for k in list(temp_vars.keys()):
            del temp_vars[k]

    def _get_temp_data_args(self, temp_vars):
        result = [temp_vars[v] for v in self.index_function_args]
        return result

    def _reorder_particles(self, depth, child_count_prefix_sum, offsets_parent,
                           pbounds_parent,
                           seg_flag, csum_nodes_prev, temp_vars):
        # Scan

        args = [('__global ' + ctype + ' *' + v) for v, ctype in
                zip(self.index_function_args, self.index_function_arg_ctypes)]
        args += [(ctype + ' ' + v) for v, ctype in
                 zip(self.index_function_consts,
                     self.index_function_const_ctypes)]
        args = ', '.join(args)

        particle_kernel = _get_particle_kernel(self.ctx, self.k,
                                               args, self.index_code)
        args = [seg_flag.dev, child_count_prefix_sum.dev]
        args += [x.dev for x in self.get_data_args()]
        args += self.get_index_constants(depth)
        particle_kernel(*args)

        # Reorder particles
        reorder_particles = self.main_helper.get_kernel(
            'reorder_particles', k=self.k,
            data_vars=tuple(self.index_function_args),
            data_var_ctypes=tuple(self.index_function_arg_ctypes),
            const_vars=tuple(self.index_function_consts),
            const_var_ctypes=tuple(self.index_function_const_ctypes),
            index_code=self.index_code
        )

        args = [self.pids.dev, self.cids.dev,
                seg_flag.dev,
                pbounds_parent.dev, offsets_parent.dev,
                child_count_prefix_sum.dev,
                temp_vars['pids'].dev, temp_vars['cids'].dev]
        args += [x.dev for x in self.get_data_args()]
        args += [x.dev for x in self._get_temp_data_args(temp_vars)]
        args += self.get_index_constants(depth)
        args += [np.uint32(csum_nodes_prev)]

        reorder_particles(*args)
        self._exchange_temp_vars(temp_vars)

    def _merge_layers(self, offsets_temp, pbounds_temp):
        curr_offset = 0
        total_nodes = 0

        for i in range(self.depth + 1):
            total_nodes += self.num_nodes[i]

        self.offsets = Array(np.int32, n=total_nodes, backend='opencl')
        self.pbounds = Array(cl.cltypes.uint2, n=total_nodes, backend='opencl')

        append_layer = self.main_helper.get_kernel('append_layer')

        self.total_nodes = total_nodes
        for i in range(self.depth + 1):
            append_layer(
                offsets_temp[i].dev, pbounds_temp[i].dev,
                self.offsets.dev, self.pbounds.dev,
                np.int32(curr_offset), np.uint8(i == self.depth)
            )
            curr_offset += self.num_nodes[i]

    def _update_node_data(self, offsets_prev, pbounds_prev, offsets, pbounds,
                          seg_flag, child_count_prefix_sum, csum_nodes,
                          csum_nodes_next, n):
        """Update node data and return number of children which are leaves."""

        # Update particle-related data of children
        set_node_data = self.main_helper.get_kernel("set_node_data", k=self.k)
        set_node_data(offsets_prev.dev, pbounds_prev.dev,
                      offsets.dev, pbounds.dev,
                      seg_flag.dev, child_count_prefix_sum.dev,
                      np.uint32(csum_nodes),
                      np.uint32(n))

        # Set children offsets
        leaf_count = Array(np.uint32, n=1, backend='opencl')
        set_offsets = _get_set_offset_kernel(self.ctx, self.k, self.leaf_size)
        set_offsets(pbounds.dev, offsets.dev, leaf_count.dev,
                    np.uint32(csum_nodes_next))
        return leaf_count.dev[0].get()

    def _build_tree(self, fixed_depth=None):
        # We build the tree one layer at a time. We stop building new
        # layers after either all the
        # nodes are leaves or after reaching the target depth (fixed_depth).
        # At this point, the information for each layer is segmented / not
        # contiguous in memory, and so we run a merge_layers procedure to
        # move the data for all layers into a single array.
        #
        # The procedure for building each layer can be split up as follows
        # 1) Determine which child each particle is going to belong to in the
        #    next layer
        # 2) Perform a kind of segmented scan over this. This gives us the
        #    new order of the particles so that consecutive particles lie in
        #    the same child
        # 3) Reorder the particles based on this order
        # 4) Create a new layer and set the node data for the new layer. We
        #    get to know which particles belong to each node directly from the
        #    results of step 2
        # 5) Set the predicted offsets of the children of the nodes in the
        #    new layer. If a node has fewer than leaf_size particles, it's a
        #    leaf. A kind of prefix sum over this directly let's us know the
        #    predicted offsets.
        # Rinse and repeat for building more layers.
        #
        # Note that after building the last layer, the predicted offsets for
        # the children might not be correctly since we're not going to build
        # more layers. The _merge_layers procedure sets the offsets in the
        # last layer to -1 to correct this.

        num_leaves_here = 0
        n = self.n
        temp_vars = {}

        self.depth = 0
        self.num_nodes = [1]

        # Cumulative sum of nodes in the previous layers
        csum_nodes_prev = 0
        csum_nodes = 1

        # Initialize temporary data (but persistent across layers)
        self._create_temp_vars(temp_vars)

        child_count_prefix_sum = Array(get_vector_dtype('uint', self.k),
                                       n=n, backend='opencl')

        seg_flag = Array(cl.cltypes.char, n=n, backend='opencl')
        seg_flag.fill(0)
        seg_flag.dev[0] = 1

        offsets_temp = [Array(np.int32, n=1, backend='opencl')]
        offsets_temp[-1].fill(1)

        pbounds_temp = [Array(cl.cltypes.uint2, n=1, backend='opencl')]
        pbounds_temp[-1].dev[0].set(cl.cltypes.make_uint2(0, n))

        # FIXME: Depths above 20 possible and feasible for binary / quad trees
        loop_lim = min(fixed_depth, 20)

        for depth in range(1, loop_lim):
            num_nodes = self.k * (self.num_nodes[-1] - num_leaves_here)
            if num_nodes == 0:
                break
            else:
                self.depth += 1
            self.num_nodes.append(num_nodes)

            # Allocate new layer
            offsets_temp.append(Array(np.int32, n=self.num_nodes[-1],
                                      backend='opencl'))
            pbounds_temp.append(Array(cl.cltypes.uint2,
                                      n=self.num_nodes[-1], backend='opencl'))

            # Generate particle index and reorder the particles
            self._reorder_particles(depth, child_count_prefix_sum,
                                    offsets_temp[-2],
                                    pbounds_temp[-2], seg_flag,
                                    csum_nodes_prev,
                                    temp_vars)

            num_leaves_here = self._update_node_data(
                offsets_temp[-2], pbounds_temp[-2],
                offsets_temp[-1], pbounds_temp[-1],
                seg_flag, child_count_prefix_sum,
                csum_nodes, csum_nodes + self.num_nodes[-1], n
            )

            csum_nodes_prev = csum_nodes
            csum_nodes += self.num_nodes[-1]

        self._merge_layers(offsets_temp, pbounds_temp)
        self._clean_temp_vars(temp_vars)

    ###########################################################################
    # Misc
    ###########################################################################

    def _get_unique_cids_and_count(self):
        n = self.n
        self.unique_cids = Array(np.uint32, n=n, backend='opencl')
        self.unique_cids_map = Array(np.uint32, n=n, backend='opencl')
        uniq_count = Array(np.uint32, n=1, backend='opencl')
        unique_cids_kernel = _get_unique_cids_kernel(self.ctx)
        unique_cids_kernel(self.cids.dev, self.unique_cids_map.dev,
                           self.unique_cids.dev, uniq_count.dev)
        self.unique_cid_count = uniq_count.dev[0].get()

    def get_leaves(self):
        leaves = Array(np.uint32, n=self.offsets.dev.shape[0], backend='opencl')
        num_leaves = Array(np.uint32, n=1, backend='opencl')
        leaves_kernel = _get_leaves_kernel(self.ctx, self.leaf_size)
        leaves_kernel(self.offsets.dev, self.pbounds.dev,
                      leaves.dev, num_leaves.dev)

        num_leaves = num_leaves.dev[0].get()
        return leaves.dev[:num_leaves], num_leaves

    def _sort(self):
        """Set tree as being sorted

        The particle array needs to be aligned by the caller!
        """
        if not self.sorted:
            self.sorted = 1

    ###########################################################################
    # Tree API
    ###########################################################################
    def allocate_node_prop(self, dtype):
        return Array(dtype, n=self.total_nodes, backend='opencl')

    def allocate_leaf_prop(self, dtype):
        return Array(dtype, n=int(self.unique_cid_count), backend='opencl')

    def get_preamble(self):
        if self.sorted:
            return "#define PID(idx) (idx)"
        else:
            return "#define PID(idx) (pids[idx])"

    def get_leaf_size_partitions(self, group_min, group_max):
        """Partition leaves based on leaf size

        Parameters
        ----------
        group_min
            Minimum leaf size
        group_max
            Maximum leaf size
        Returns
        -------
        groups : Array
            An array which contains the cell ids of leaves
            with leaf size > group_min and leaf size <= group_max
        group_count : int
            The number of leaves which satisfy the given condition
            on the leaf size
        """
        groups = Array(np.uint32, n=int(self.unique_cid_count),
                       backend='opencl')
        group_count = Array(np.uint32, n=1, backend='opencl')

        get_cid_groups = _get_cid_groups_kernel(self.ctx)
        get_cid_groups(self.unique_cids.dev[:self.unique_cid_count],
                       self.pbounds.dev, groups.dev, group_count.dev,
                       np.int32(group_min), np.int32(group_max))
        result = groups, int(group_count.dev[0].get())
        return result

    def tree_bottom_up(self, args, setup, leaf_operation, node_operation,
                       output_expr, preamble=""):
        return tree_bottom_up(self.ctx, args, setup, leaf_operation,
                              node_operation, output_expr, preamble)

    def leaf_tree_traverse(self, args, setup, node_operation, leaf_operation,
                           output_expr, common_operation="", preamble=""):
        """
        Traverse this (source) tree. One thread for each leaf of
        destination tree.
        """

        return leaf_tree_traverse(self.ctx, self.k, args, setup,
                                  node_operation, leaf_operation,
                                  output_expr, common_operation, preamble)

    def point_tree_traverse(self, args, setup, node_operation, leaf_operation,
                            output_expr, common_operation="", preamble=""):
        """
        Traverse this (source) tree. One thread for each particle of
        destination tree.
        """

        return point_tree_traverse(self.ctx, self.k, args, setup,
                                   node_operation, leaf_operation,
                                   output_expr, common_operation, preamble)



"""2-d Riemann problem (3 hours)

four shock waves' interaction at the centerof the domain
"""

# numpy and standard imports
import numpy

# pysph imports
from pysph.base.utils import get_particle_array as gpa
from pysph.base.nnps import DomainManager
from pysph.solver.application import Application
from pysph.sph.scheme import GSPHScheme, SchemeChooser, ADKEScheme, GasDScheme
from pysph.sph.wc.crksph import CRKSPHScheme
from pysph.sph.gas_dynamics.psph import PSPHScheme
from pysph.sph.gas_dynamics.tsph import TSPHScheme
from pysph.examples.gas_dynamics.riemann_2d_config import R2DConfig

# current case from the al possible unique cases
case = 3

# config for current case
config = R2DConfig(case)
gamma = 1.4
gamma1 = gamma - 1
kernel_factor = 1.5
dt = 1e-4
dim = 2


class Riemann2D(Application):
    def initialize(self):
        # square domain
        self.dt = dt
        self.tf = config.endtime

    def add_user_options(self, group):
        group.add_argument(
            "--dscheme", choices=["constant_mass", "constant_volume"],
            dest="dscheme", default="constant_volume",
            help="spatial discretization scheme, one of constant_mass"
            "or constant_volume"
        )
        group.add_argument(
            "--nparticles", action="store", type=int,
            dest="nparticles", default=200
        )

    def consume_user_options(self):
        self.nx = self.options.nparticles
        self.ny = self.nx
        self.dx = (config.xmax - config.xmin) / self.nx

        # discretization function
        if self.options.dscheme == "constant_volume":
            self.dfunction = self.create_particles_constant_volume
        elif self.options.dscheme == "constant_mass":
            self.dfunction = self.create_particles_constant_mass

    def create_particles_constant_volume(self):
        dx = self.dx
        dx2 = dx * 0.5
        vol = dx * dx

        xmin = config.xmin
        ymin = config.ymin
        xmax = config.xmax
        ymax = config.ymax
        xmid = config.xmid
        ymid = config.ymid

        rho1, u1, v1, p1 = config.rho1, config.u1, config.v1, config.p1
        rho2, u2, v2, p2 = config.rho2, config.u2, config.v2, config.p2
        rho3, u3, v3, p3 = config.rho3, config.u3, config.v3, config.p3
        rho4, u4, v4, p4 = config.rho4, config.u4, config.v4, config.p4
        x, y = numpy.mgrid[xmin+dx2:xmax:dx, ymin+dx2:ymax:dx]

        x = x.ravel()
        y = y.ravel()

        u = numpy.zeros_like(x)
        v = numpy.zeros_like(x)

        # density and mass
        rho = numpy.ones_like(x)
        p = numpy.ones_like(x)

        for i in range(x.size):
            if x[i] <= xmid:
                if y[i] <= ymid:  # w3
                    rho[i] = rho3
                    p[i] = p3
                    u[i] = u3
                    v[i] = v3
                else:            # w2
                    rho[i] = rho2
                    p[i] = p2
                    u[i] = u2
                    v[i] = v2
            else:
                if y[i] <= ymid:  # w4
                    rho[i] = rho4
                    p[i] = p4
                    u[i] = u4
                    v[i] = v4
                else:            # w1
                    rho[i] = rho1
                    p[i] = p1
                    u[i] = u1
                    v[i] = v1

        # thermal energy
        e = p/(gamma1 * rho)

        # mass
        m = vol * rho

        # smoothing length
        h = kernel_factor * (m/rho)**(1./dim)

        # create the particle array
        pa = gpa(name='fluid', x=x, y=y, m=m, rho=rho, h=h,
                 u=u, v=v, p=p, e=e, h0=h.copy())

        return pa

    def create_particles_constant_mass(self):
        dx = self.dx

        xmin = config.xmin
        ymin = config.ymin
        xmax = config.xmax
        ymax = config.ymax
        xmid = config.xmid
        ymid = config.ymid

        rho_max = config.rho_max
        nb4 = self.nx/4
        dx0 = (xmax - xmid)/nb4
        vol0 = dx0 * dx0
        m0 = rho_max * vol0

        # first quadrant
        vol1 = config.rho_max/config.rho1 * vol0
        dx = numpy.sqrt(vol1)
        dxb2 = 0.5 * dx
        x1, y1 = numpy.mgrid[xmid+dxb2:xmax:dx, ymid+dxb2:ymax:dx]
        x1 = x1.ravel()
        y1 = y1.ravel()

        u1 = numpy.ones_like(x1) * config.u1
        v1 = numpy.zeros_like(x1) * config.v1
        rho1 = numpy.ones_like(x1) * config.rho1
        p1 = numpy.ones_like(x1) * config.p1
        m1 = numpy.ones_like(x1) * m0
        h1 = numpy.ones_like(x1) * kernel_factor * (m1/rho1)**(0.5)

        # second quadrant
        vol2 = config.rho_max/config.rho2 * vol0
        dx = numpy.sqrt(vol2)
        dxb2 = 0.5 * dx
        x2, y2 = numpy.mgrid[xmid-dxb2:xmin:-dx, ymid+dxb2:ymax:dx]
        x2 = x2.ravel()
        y2 = y2.ravel()
        u2 = numpy.ones_like(x2) * config.u2
        v2 = numpy.ones_like(x2) * config.v2
        rho2 = numpy.ones_like(x2) * config.rho2
        p2 = numpy.ones_like(x2) * config.p2
        m2 = numpy.ones_like(x2) * m0
        h2 = numpy.ones_like(x2) * kernel_factor * (m2/rho2)**(0.5)

        # third quadrant
        vol3 = config.rho_max/config.rho3 * vol0
        dx = numpy.sqrt(vol3)
        dxb2 = 0.5 * dx
        x3, y3 = numpy.mgrid[xmid-dxb2:xmin:-dx, ymid-dxb2:ymin:-dx]
        x3 = x3.ravel()
        y3 = y3.ravel()

        u3 = numpy.ones_like(x3) * config.u3
        v3 = numpy.ones_like(x3) * config.v3
        rho3 = numpy.ones_like(x3) * config.rho3
        p3 = numpy.ones_like(x3) * config.p3
        m3 = numpy.ones_like(x3) * m0
        h3 = numpy.ones_like(x3) * kernel_factor * (m3/rho3)**(0.5)

        # fourth quadrant
        vol4 = config.rho_max/config.rho4 * vol0
        dx = numpy.sqrt(vol4)
        dxb2 = 0.5 * dx
        x4, y4 = numpy.mgrid[xmid+dxb2:xmax:dx, ymid-dxb2:ymin:-dx]
        x4 = x4.ravel()
        y4 = y4.ravel()

        u4 = numpy.ones_like(x4) * config.u4
        v4 = numpy.ones_like(x4) * config.v4
        rho4 = numpy.ones_like(x4) * config.rho4
        p4 = numpy.ones_like(x4) * config.p4
        m4 = numpy.ones_like(x4) * m0
        h4 = numpy.ones_like(x4) * kernel_factor * (m4/rho4)**(0.5)

        # concatenate the arrays
        x = numpy.concatenate([x1, x2, x3, x4])
        y = numpy.concatenate([y1, y2, y3, y4])
        p = numpy.concatenate([p1, p2, p3, p4])
        u = numpy.concatenate([u1, u2, u3, u4])
        v = numpy.concatenate([v1, v2, v3, v4])
        h = numpy.concatenate([h1, h2, h3, h4])
        m = numpy.concatenate([m1, m2, m3, m4])
        rho = numpy.concatenate([rho1, rho2, rho3, rho4])

        # derived variables
        e = p/((gamma-1.0) * rho)

        # create the particle array
        pa = gpa(
            name='fluid', x=x, y=y, m=m, rho=rho, h=h, u=u, v=v, p=p, e=e,
            h0=h.copy()
        )

        return pa

    def create_particles(self):
        fluid = self.dfunction()
        self.scheme.setup_properties([fluid])

        return [fluid]

    def create_domain(self):
        return DomainManager(
            xmin=config.xmin, xmax=config.xmax,
            ymin=config.ymin, ymax=config.ymax,
            mirror_in_x=True, mirror_in_y=True
        )

    def create_scheme(self):
        gsph = GSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            kernel_factor=1.5,
            g1=0.25, g2=0.5, rsolver=2, interpolation=1, monotonicity=1,
            interface_zero=True, hybrid=False, blend_alpha=2.0,
            niter=40, tol=1e-6, has_ghosts=True
        )

        adke = ADKEScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            alpha=1, beta=1.0, k=1.0, eps=0.5, g1=0.2, g2=0.4,
            has_ghosts=True)

        crksph = CRKSPHScheme(
            fluids=['fluid'], dim=dim, rho0=0, c0=0, nu=0, h0=0, p0=0,
            gamma=gamma, cl=2, has_ghosts=True
        )

        mpm = GasDScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            kernel_factor=1.2, alpha1=1.0, alpha2=0.1,
            beta=2.0, update_alpha1=True, update_alpha2=True,
            has_ghosts=True
        )

        psph = PSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            hfact=1.2
        )

        tsph = TSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            hfact=1.2
        )

        s = SchemeChooser(
            default='gsph', gsph=gsph, adke=adke, crksph=crksph, mpm=mpm,
            psph=psph, tsph=tsph
        )
        return s

    def configure_scheme(self):
        s = self.scheme
        if self.options.scheme == 'gsph':
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=50)
        elif self.options.scheme == 'adke':
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=50)
        elif self.options.scheme == 'crksph':
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=50)
        elif self.options.scheme == 'mpm':
            s.configure(kernel_factor=kernel_factor)
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=50)
        elif self.options.scheme in ['tsph', 'psph']:
            s.configure(hfact=kernel_factor)
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=50)

    def post_process(self):
        if len(self.output_files) < 1 or self.rank > 0:
            return
        try:
            import matplotlib
            matplotlib.use('Agg')
            from matplotlib import pyplot
        except ImportError:
            print("Post processing requires matplotlib.")
            return
        from pysph.solver.utils import load
        import os
        outfile = self.output_files[-1]
        data = load(outfile)
        pa = data['arrays']['fluid']
        x = pa.x
        y = pa.y
        pyplot.scatter(x, y, s=1)
        pyplot.xlim((0.1, 0.6))
        pyplot.ylim((0.1, 0.6))
        fig = os.path.join(self.output_dir, "positions.png")
        pyplot.savefig(fig, dpi=300)
        pyplot.close('all')


if __name__ == "__main__":
    app = Riemann2D()
    app.run()
    app.post_process()



from pysph.examples.flow_past_cylinder_2d import WindTunnel
from pysph.base.utils import get_particle_array
from math import sin, cos, pi
import numpy as np
import tempfile
import os

# fluid mechanical/numerical parameters
rho = 1000
umax = 1.0
c0 = 10 * umax
p0 = rho * c0 * c0
use_coords = True

# creating the files and coordinates of the cylinder surface for demonstration
xc, yc = [], []
cyl_file = os.path.join(tempfile.gettempdir(), 'cylinder.txt')
print(cyl_file)
fp = open(cyl_file, 'w')
for i in range(0, 100):
    _x = cos(2 * pi * i / 100) + 5.0
    _y = sin(2 * pi * i / 100)
    xc.append(_x)
    yc.append(_y)
    fp.write('%.3f %.3f\n' % (_x, _y))
fp.close()


class FPCWithPackedCylinder(WindTunnel):
    def _get_packed_points(self):
        '''
        returns
        xs, ys, zs, xf, yf, zf
        '''
        from pysph.tools.geometry import (
            get_packed_2d_particles_from_surface_coordinates,
            get_packed_2d_particles_from_surface_file)
        folder = self.output_dir
        dx = self.dx
        if use_coords:
            return get_packed_2d_particles_from_surface_coordinates(
                self.add_user_options, folder, dx, x=np.array(xc),
                y=np.array(yc), shift=True)
        else:
            return get_packed_2d_particles_from_surface_file(
                self.add_user_options, folder, dx, filename=cyl_file,
                shift=True)

    def _create_solid(self):
        xs, ys, zs, xf, yf, zf = self._get_packed_points()
        dx = self.dx
        h0 = self.h
        volume = dx*dx
        solid = get_particle_array(
            name='solid', x=xs-dx/2, y=ys,
            m=volume*rho, rho=rho, h=h0, V=1.0/volume)
        return solid

    def _create_fluid(self):
        from pysph.tools.geometry import create_fluid_around_packing
        xs, ys, zs, xf, yf, zf = self._get_packed_points()
        dx = self.dx
        h0 = self.h
        volume = dx*dx
        L = self.Lt
        B = self.Wt * 2.0

        fluid = create_fluid_around_packing(
            dx, xf-dx/2, yf, L, B, m=volume*rho, rho=rho, h=h0, V=1.0/volume,
            u=umax, p=0.0, uhat=umax)

        return fluid

    def create_particles(self):
        dx = self.dx
        fluid = self._create_fluid()
        solid = self._create_solid()
        outlet = self._create_outlet()
        inlet = self._create_inlet()
        wall = self._create_wall()

        ghost_inlet = self.iom.create_ghost(inlet, inlet=True)
        ghost_outlet = self.iom.create_ghost(outlet, inlet=False)

        particles = [fluid, inlet, outlet, solid, wall]
        if ghost_inlet:
            particles.append(ghost_inlet)
        if ghost_outlet:
            particles.append(ghost_outlet)

        self.scheme.setup_properties(particles)
        self._set_wall_normal(wall)

        if self.io_method == 'hybrid':
            fluid.uag[:] = 1.0
            fluid.uta[:] = 1.0
            outlet.uta[:] = 1.0

        return particles


if __name__ == '__main__':
    app = FPCWithPackedCylinder()
    app.run()
    app.post_process(app.info_filename)



from math import sqrt
import numpy as np
import os

from pysph.sph.surface_tension import get_surface_tension_equations

from pysph.tools.geometry import get_2d_block, remove_overlap_particles
from pysph.base.utils import get_particle_array
from pysph.base.kernels import CubicSpline, QuinticSpline
from pysph.solver.application import Application
from pysph.solver.solver import Solver

from pysph.sph.integrator_step import TransportVelocityStep
from pysph.sph.integrator import PECIntegrator

from pysph.base.nnps import DomainManager
from pysph.solver.utils import iter_output

dim = 2
Lx = 1.0
Ly = 1.0

nu = 0.05
sigma = 1.0
factor1 = 0.8
factor2 = 1 / factor1
rho0 = 1.0

c0 = 20.0
gamma = 1.4
R = 287.1

p0 = c0**2 * rho0

nx = 120
dx = Lx / nx
volume = dx * dx

tf = 0.5

hdx = 1.5
h0 = hdx * dx
epsilon = 0.01 / h0

r0 = 0.05
v0 = 10.0


dt1 = 0.25*np.sqrt(rho0*h0*h0*h0/(2.0*np.pi*sigma))

dt2 = 0.25*h0/(c0+v0)

dt3 = 0.125*rho0*h0*h0/nu

dt = 0.9*min(dt1, dt2, dt3)


def r(x, y):
    return x*x + y*y


class MultiPhase(Application):

    def create_particles(self):
        c0 = 20
        hdx = 1.5
        h0 = hdx * dx
        if self.options.scheme == 'adami_stress':
            c0 = 10
            hdx = 1.0
            h0 = hdx*dx
            dt1 = 0.25*np.sqrt(rho0*h0*h0*h0/(2.0*np.pi*sigma))
            dt2 = 0.25*h0/(c0+v0)
            dt3 = 0.125*rho0*h0*h0/nu
            dt = 0.9*min(dt1, dt2, dt3)

        fluid_x, fluid_y = get_2d_block(
            dx=dx, length=Lx, height=Ly, center=np.array([0., 0.]))
        rho_fluid = np.ones_like(fluid_x) * rho0
        m_fluid = rho_fluid * volume
        h_fluid = np.ones_like(fluid_x) * h0
        cs_fluid = np.ones_like(fluid_x) * c0
        wall_x, wall_y = get_2d_block(dx=dx, length=Lx+6*dx, height=Ly+6*dx,
                                      center=np.array([0., 0.]))
        rho_wall = np.ones_like(wall_x) * rho0
        m_wall = rho_wall * volume
        h_wall = np.ones_like(wall_x) * h0
        cs_wall = np.ones_like(wall_x) * c0
        additional_props = ['V', 'color', 'scolor', 'cx', 'cy', 'cz',
                            'cx2', 'cy2', 'cz2', 'nx', 'ny', 'nz', 'ddelta',
                            'uhat', 'vhat', 'what', 'auhat', 'avhat', 'awhat',
                            'ax', 'ay', 'az', 'wij', 'vmag2', 'N', 'wij_sum',
                            'rho0', 'u0', 'v0', 'w0', 'x0', 'y0', 'z0',
                            'kappa', 'arho', 'nu', 'wg', 'ug', 'vg',
                            'pi00', 'pi01', 'pi02', 'pi10', 'pi11', 'pi12',
                            'pi20', 'pi21', 'pi22', 'alpha']
        consts = {'max_ddelta': np.zeros(1, dtype=float)}

        fluid = get_particle_array(
            name='fluid', x=fluid_x, y=fluid_y, h=h_fluid, m=m_fluid,
            rho=rho_fluid, cs=cs_fluid, additional_props=additional_props,
            constants=consts)
        for i in range(len(fluid.x)):
            if (fluid.x[i]*fluid.x[i] + fluid.y[i]*fluid.y[i]) < 0.04:
                fluid.color[i] = 1.0
            else:
                fluid.color[i] = 0.0
        fluid.alpha[:] = sigma
        wall = get_particle_array(
            name='wall', x=wall_x, y=wall_y, h=h_wall, m=m_wall,
            rho=rho_wall, cs=cs_wall, additional_props=additional_props)
        wall.color[:] = 0.0
        remove_overlap_particles(wall, fluid, dx_solid=dx, dim=2)
        fluid.add_output_arrays(['V', 'color', 'cx', 'cy', 'nx', 'ny',
                                 'ddelta', 'kappa', 'N', 'scolor', 'p'])
        wall.add_output_arrays(['V', 'color', 'cx', 'cy', 'nx', 'ny', 'ddelta',
                                'kappa', 'N', 'scolor', 'p'])
        for i in range(len(fluid.x)):
            R = sqrt(r(fluid.x[i], fluid.y[i]) + 0.0001*fluid.h[i]*fluid.h[i])
            f = np.exp(-R/r0)/r0
            fluid.u[i] = v0*fluid.x[i]*(1.0-(fluid.y[i]*fluid.y[i])/(r0*R))*f
            fluid.v[i] = -v0*fluid.y[i]*(1.0-(fluid.x[i]*fluid.x[i])/(r0*R))*f
        fluid.nu[:] = nu

        return [fluid, wall]

    def create_solver(self):
        kernel = QuinticSpline(dim=2)
        integrator = PECIntegrator(fluid=TransportVelocityStep())
        solver = Solver(
            kernel=kernel, dim=dim, integrator=integrator,
            dt=dt, tf=tf, adaptive_timestep=False,
            output_at_times=[0., 0.08, 0.16, 0.26])
        return solver

    def add_user_options(self, group):
        choices = ['morris', 'tvf', 'adami_stress', 'adami', 'shadloo']
        group.add_argument(
            "--scheme", action="store", dest='scheme', default='morris',
            choices=choices,
            help='Specify scheme to use among %s' % choices
        )

    def create_equations(self):
        return get_surface_tension_equations(['fluid'], ['wall'],
                                             self.options.scheme, rho0, p0, c0,
                                             0,  factor1, factor2, nu, sigma,
                                             2, epsilon, gamma, real=True)

    def post_process(self):
        try:
            import matplotlib
            matplotlib.use('Agg')
            import matplotlib.pyplot as plt
        except ImportError:
            print("Post processing requires Matplotlib")
            return
        from pysph.solver.utils import load
        files = self.output_files
        amat = []
        t = []
        centerx = []
        centery = []
        velx = []
        vely = []

        for f in files:
            data = load(f)
            pa = data['arrays']['fluid']
            t.append(data['solver_data']['t'])
            x = pa.x
            y = pa.y
            u = pa.u
            v = pa.v
            color = pa.color
            length = len(color)
            min_x = 0.0
            max_x = 0.0
            cx = 0
            cy = 0
            vx = 0
            vy = 0
            count = 0
            for i in range(length):
                if color[i] == 1:
                    if x[i] < min_x:
                        min_x = x[i]
                    if x[i] > max_x:
                        max_x = x[i]
                    if x[i] > 0 and y[i] > 0:
                        cx += x[i]
                        cy += y[i]
                        vx += u[i]
                        vy += v[i]
                        count += 1
            amat.append(0.5*(max_x - min_x))
            centerx.append(cx/count)
            centery.append(cy/count)
            velx.append(vx/count)
            vely.append(vy/count)
        fname = os.path.join(self.output_dir, 'results.npz')
        np.savez(fname, t=t, semimajor=amat, centerx=centerx, centery=centery,
                 velx=velx, vely=vely)
        plt.plot(t, amat)
        fig = os.path.join(self.output_dir, 'semimajorvst.png')
        plt.savefig(fig)
        plt.close()
        plt.plot(t, centerx, label='x position')
        plt.plot(t, centery, label='y position')
        plt.legend()
        fig1 = os.path.join(self.output_dir, 'centerofmassposvst')
        plt.savefig(fig1)
        plt.close()
        plt.plot(t, velx, label='x velocity')
        plt.plot(t, vely, label='y velocity')
        plt.legend()
        fig2 = os.path.join(self.output_dir, 'centerofmassvelvst')
        plt.savefig(fig2)
        plt.close()


if __name__ == '__main__':
    app = MultiPhase()
    app.run()
    app.post_process()
